/*
 *  This program reads a CICY list and removes redundancies due to ineffective splittings and identities.
 *  It can also remove redundancies due to permutations of rows and columns. However, for large matrices and large lists,
 *  it is faster to use the Mathematica code "2ndfilter_perm.m" for that.
 *
 *  Created by Alexander Haupt on 12/05/2010.
 *  Copyright 2013 Alexander Haupt. All rights reserved.
 *
 */

#if defined(_WIN32) || defined(_WIN64)
  #error not tested on windows
#endif

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <time.h>
#include <set>
#include <vector>
#include <algorithm>
#include <sys/stat.h>

#if defined(__APPLE__) || defined(__MACH__) || defined(__OSX__)
  #include <malloc/malloc.h>
#else
  #include <malloc.h>
#endif

#include "euler.h"
#include "mu.h"
#include "qmat_cls.h"
#include "is_already_there.h"

using namespace std;
using namespace __gnu_cxx;

int n;

void proc_cicy_list(const int offset1, const int offset2, const char filename[]);
static inline void write_output1(FILE * fp, const int counter, const int m, const int K, const register int n_r[], const register int q[], const int eulernumber);
inline void debug_output(const int m, const int K, const int n_r[], const int q[]);
inline void init_num_P(const int m, const int n_r[], int num_P[]);
inline int * init_array(const int dim, const int val);
inline void init_val(const int dim, int array[], const int val);
inline int sum_q_column(const int a, const int m, const int K, const int q[]);
inline int is_0fold_id(const int m, const int K, const int n_r[], const int q[], const int num_P[]);
inline int is_1fold_id(const int m, const int K, const int n_r[], const int q[]);
inline int is_2fold_id(const int m, const int K, const int n_r[], const int q[], const int num_P[]);
inline int check_T2_redundancy(const int m, const int K, const register int n_r[], const register int q[], const int num_P[]);
inline int check_K3_redundancy(const int m, const int K, const register int n_r[], const register int q[], const int num_P[]);
inline int is_contractible_row(const int row, const int K, const register int q[]);
inline conf* contract(const int old_m, const int old_K, const int old_n_r[], const int old_q[], const int old_eulernumber);
inline int build_and_calc_subdetvar(const int my_n, const int old_m, const int old_K, const int old_q[], const int tmp_K, const int i, const int new_n_r[], const int col1[], const int col2[], const int col3[], const int col4[], int * empty);
inline int calc_nu_comp(const int old_m, const int old_K, const int old_q[], const int tmp_K, const int i, const int new_n_r[], const int col1[], const int col2[], const int col3[], const int col4[], const int col5[]);
inline conf* apply_identities(const int old_m, const int old_K, const int old_n_r[], const int old_q[], const int old_eulernumber);


int main (int argc, const char * argv[]) {
	if (argc != 5) {
		// input: n for CICY n-fold
		// offset1: 1st #offset1 matrices are not checked with "is_already_there"
		// offset2: matrices numbers > offset2 are not checked with "is_already_there"
		// in other words; the only matrix numbers that are checked with "is_already_there" lie in the interval [(offset1+1),offset2]
		printf("Usage: %s n offset1 offset2 filename\n", argv[0]);
	}
	else {
		n = atoi(argv[1]);
		proc_cicy_list(atoi(argv[2]), atoi(argv[3]), argv[4]);
	}
	
	return 0;
}

void proc_cicy_list(const int offset1, const int offset2, const char filename[]) {
	int i, linecounter = 0, counter = 0;
	int m, K, eulernumber, orig_matno;
	int * n_r;
	int * q;		// q part of CICY config. matrix
	int num_P[2*(n+1)];	// array fr. 1 to (2*n+1). num_P[i] is the number of \CP^i factors in n_r. Note: n_r is ordered (ascendingly). num_P[0] is ignored
	
	FILE * fp = fopen(filename, "r");

	set<conf> conflist;
	set<conf>::iterator it;
	
	char output_filename[50]; time_t timer; 
	timer = time(NULL);
	sprintf(output_filename, "2ndf_%s", basename(filename));
	
	if (fp != NULL) {
		char line[128], strnum[2] = {0,0};
		
		while(fgets(line, sizeof line, fp) != NULL) { /* read a line */
			switch (linecounter) {
				case 0:
					if (line[0] == 'M') {
						sscanf(line, "MATRIX NUMBER :%6i\n", &orig_matno);
						linecounter++;
					}
					break;
				case 1:
					sscanf(line, "SIZE :%4i%3i\n", &m, &K);
					n_r = (int*)malloc(m*sizeof(int));
					q = (int*)malloc(m*K*sizeof(int));
					linecounter++;
					//printf("m=%i, K=%i, ", m, K);	// debug
					break;
				case 2:
					sscanf(line, "EULERNUMBER :%5i\n", &eulernumber);
					linecounter++;
					//printf("eulernumber=%i\n", eulernumber);	// debug
					break;
				default:
					if (linecounter <= (m+2)) {
						n_r[linecounter-3] = 0;
						for (i = 0; i < K; i++) {
							strnum[0] = line[(2*i)+4];
							q[i+((linecounter-3)*K)] = atoi(strnum);
							n_r[linecounter-3] += q[i+((linecounter-3)*K)];
						}
						n_r[linecounter-3]--;
						// only keep CICYs which are embedded in ambient spaces that don't contain P1s
						/*if ( (linecounter == 3) && (n_r[linecounter-3] == 1) ) {
							linecounter = 0;
							continue;
						}*/
						if (linecounter == (m+2)) {
							time(&timer); printf("%i:\t%s", ++counter, ctime(&timer)); fflush(stdout);
							if ( (counter > offset1) && (counter <= offset2) ) {
								if (is_already_there(&conflist, m, K, n_r, q) == conflist.end()) {
									add_conf(&conflist, m, K, n_r, q, eulernumber);
								}
							}
							else {
								add_conf(&conflist, m, K, n_r, q, eulernumber);
							}
						}
						linecounter++;
					}
					else {
						linecounter = 0;
					}
					break;
			}
		}
		fclose(fp);
	}
	else {
		perror(filename); /* why didn't the file open? */
	}

	printf("counter: %i\n", (int)conflist.size());
	conf *new_conf, *old_conf; int to_be_removed; counter = 0;
	for ( it=conflist.begin(); it != conflist.end(); ) { //it++ ) {
		time(&timer); printf("%i:\t%s", ++counter, ctime(&timer)); fflush(stdout);
		to_be_removed = false;
		
		memset(num_P, 0, 2*(n+1)*sizeof(int));
		init_num_P((*it).m, (*it).n_r, num_P);

		if ( is_0fold_id((*it).m, (*it).K, (*it).n_r, (*it).q, num_P) ) {
			to_be_removed = true;
		}

		if (to_be_removed == false) {
			new_conf = contract((*it).m, (*it).K, (*it).n_r, (*it).q, (*it).eulernumber);
			
			if (new_conf != NULL) {
				old_conf = new_conf;
				new_conf = apply_identities(new_conf->m, new_conf->K, new_conf->n_r, new_conf->q, new_conf->eulernumber);
				if (new_conf == NULL) {new_conf = old_conf;}
			}
			else {
				new_conf = apply_identities((*it).m, (*it).K, (*it).n_r, (*it).q, (*it).eulernumber);
			}
			
			if (new_conf != NULL) {
				to_be_removed = true;
				
				memset(num_P, 0, 2*(n+1)*sizeof(int));
				init_num_P(new_conf->m, new_conf->n_r, num_P);
				if ( !is_0fold_id(new_conf->m, new_conf->K, new_conf->n_r, new_conf->q, num_P) &&
					 (is_already_there(&conflist, new_conf->m, new_conf->K, new_conf->n_r, new_conf->q) == conflist.end()) ) {
					add_conf(&conflist, new_conf->m, new_conf->K, new_conf->n_r, new_conf->q, new_conf->eulernumber);
				}
				free(new_conf); new_conf = NULL;
			}
		}
		
		// now do the actual erase and be careful with iterator
		if (to_be_removed == true) conflist.erase(it++);
		else ++it;	
	}
	
	FILE * fp1 = fopen(output_filename, "a");
	counter = 0;
	for ( it=conflist.begin(); it != conflist.end(); it++ ) {
		write_output1(fp1, ++counter, (*it).m, (*it).K, (*it).n_r, (*it).q, (*it).eulernumber);
	}
	
	fclose(fp1);
	
	// check that writing of output file was succesful otherwise write to stderr
	int fs_err = 0;
	struct stat st;
	if (stat(output_filename, &st) != 0) {fprintf(stderr, "stat() error, errno: %i\n", errno); fs_err = 1;}
	unsigned int size = (unsigned int)st.st_size;
	printf("Output file size: %u\n", size);
	if ( (size == 0) || (fs_err == 1) ) {
		counter = 0;
		for ( it=conflist.begin(); it != conflist.end(); it++ ) {
			write_output1(stderr, ++counter, (*it).m, (*it).K, (*it).n_r, (*it).q, (*it).eulernumber);
		}
	}
}

static inline void write_output1(FILE * fp, const int counter, const int m, const int K, const register int n_r[], const register int q[], const int eulernumber) {
	int i, j;
	
	fprintf(fp, "MATRIX NUMBER :%6i\nSIZE :%4i%3i\nEULERNUMBER :%5i\n", counter, m, K, eulernumber);
	for (i=0; i < m; i++) {
		fprintf(fp, "%i |", n_r[i]);
		for (j=0; j < K; j++) {
			fprintf(fp, " %i", q[j+(i*K)]);
		}
		fprintf(fp, "\n");
	}
	fprintf(fp, "\n");
}

inline void debug_output(const int m, const int K, const int n_r[], const int q[]) {
	int i, j;
	
	for (i = 0; i < m; i++) {
		printf("%i |", n_r[i]);
		for (j = 0; j < K; j++) {
			printf(" %i", q[j+(i*K)]);
		}
		printf("\n");
	}
	printf("\n");
}

inline int * init_array(const int dim, const int val) {
	int * pointer;
	pointer = (int*)malloc(dim*sizeof(int));
	
	init_val(dim, pointer, val);
	return pointer;
}

inline void init_val(const int dim, int array[], const int val) {
	int i = 0;
	
	for (i = 0; i < dim; i++) {
		array[i] = val;
	}
}

inline void init_num_P(const int m, const int n_r[], int num_P[]) {
	int i;
	
	for (i = 0; i < m; i++) {
		num_P[n_r[i]]++;
	}
}

// returns the sum of each entry in the (a+1)-th column of q-matrix (a = 1...K)
inline int sum_q_column(const int a, const int m, const int K, const int q[]) {
	int i = 0, tmp = 0;
	
	for (i=0; i<m; i++) {
		tmp += q[(i*K) + a];
	}
	
	return tmp;
}

// returns true if [(1 1)^T | (1 1)^T], [(1 1)^T | (1 2)^T] or [2 | 2] are contained in configuration matrix 
// (as in eq. (2.3.7) of T. Huebsch, "Calabi-Yau Manifolds -- A Bestiary for Physicists")
inline int is_1fold_id(const int m, const int K, const int n_r[], const int q[]) {
	int num_P1 = 0, i = 0, j = 0, l = 0;
	
	while ( (num_P1 < m) && (n_r[num_P1] == 1) ) {
		num_P1++;
	}
	
	if (num_P1 >= 2) {
		for (i=0; i < (num_P1-1); i++) {
			for (j=i+1; j < num_P1; j++) {
				for (l=0; l < K; l++) {
					// also catch [(1 1)^T | (1 2)^T]
					if ( (q[l+(i*K)]>0) && (q[l+(j*K)]>0) && ((q[l+(i*K)]+q[l+(j*K)]) <= 3) && (sum_q_column(l,m,K,q)==(q[l+(i*K)]+q[l+(j*K)])) ) {
						return true;
					}
				}
			}
		}
	}
	
	// new code: check for [2 | 2] subconfig
	if (num_P1 == m) {
		return false;
	}
	int num_P2 = 0;
	while ( ((num_P1+num_P2) < m) && (n_r[num_P1 + num_P2] == 2)) {
		num_P2++;
	}
	if (num_P2 < 1) {
		return false;
	}
	for (i=num_P1; i < (num_P1+num_P2); i++) {
		for (j=0; j < K; j++) {
			if ( (q[j+(i*K)]==2) && (sum_q_column(j,m,K,q)==2) ) {
				return true;
			}
		}
	}
	
	return false;
}

// returns true if q-matrix has a [1 | 2] sub-configuration
inline int is_0fold_id(const int m, const int K, const int n_r[], const int q[], const int num_P[]) {
	int i = 0, j = 0;
		
	for (i=0; i < num_P[1]; i++) {
		for (j=0; j < K; j++) {
			if ( (q[j+(i*K)]==2) && (sum_q_column(j,m,K,q)==2) ) {
				return true;
			}
		}
	}
	
	return false;
}

// returns true if q-matrix has one of the following sub-configurations:
// [3|2], [4|2 2], [(2 2)|(1 1)(1 1)]^T or [(2 2)|(1 1)(1 2)]^T
inline int is_2fold_id(const int m, const int K, const int n_r[], const int q[], const int num_P[]) {
	int i1 = 0, i2 = 0, l1 = 0, l2 = 0;
	
	// [3|2]
	for (i1 = (num_P[1] + num_P[2]); i1 < (num_P[1] + num_P[2] + num_P[3]); i1++) {
		for (l1 = 0; l1 < K; l1++) {
			if ( (q[l1+(i1*K)]==2) && (sum_q_column(l1,m,K,q)==2) ) {
				return true;
			}
		}
	}
	
	// [4|2 2]
	/*for (i1 = (num_P[1] + num_P[2] + num_P[3]); i1 < (num_P[1] + num_P[2] + num_P[3] + num_P[4]); i1++) {
		for (l1 = 0; l1 < K; l1++) {
			for (l2 = 0; l2 < K; l2++) {
				if (l1 != l2) {
					if ( (q[l1+(i1*K)]==2) && (q[l2+(i1*K)]==2) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) ) {
						return true;
					}
				}
			}
		}
	}*/
	
	// [(2 2)|...]^T
	for (i1 = num_P[1]; i1 < (num_P[1] + num_P[2] - 1); i1++) {
		for (i2 = i1 + 1; i2 < (num_P[1] + num_P[2]); i2++) {
			for (l1 = 0; l1 < K; l1++) {
				for (l2 = 0; l2 < K; l2++) {
					if (l1 != l2) {
						// [(2 2)|(1 1)(1 1)]^T
						/*if ( (q[l1+(i1*K)]==1) && (q[l1+(i2*K)]==1) && 
							 (q[l2+(i1*K)]==1) && (q[l2+(i2*K)]==1) && 
							 (sum_q_column(l1,m,K,q)==2) && 
							 (sum_q_column(l2,m,K,q)==2) ) {
							return true;
						}*/
						// [(2 2)|(1 1)(1 2)]^T
						if ( (((q[l1+(i1*K)]==1) && (q[l1+(i2*K)]==1) && 
							   (q[l2+(i1*K)]==1) && (q[l2+(i2*K)]==2)) || 
							  ((q[l1+(i1*K)]==1) && (q[l1+(i2*K)]==1) && 
							   (q[l2+(i1*K)]==2) && (q[l2+(i2*K)]==1))) && 
							   (sum_q_column(l1,m,K,q)==2) && 
							   (sum_q_column(l2,m,K,q)==3) ) {
							return true;
						}
					}
				}
			}
		}
	}
	
	return false;
}

// returns TRUE if configuration matrix contains one of the following sub-configurations:
// [3|2 2] or [(2 2)|(1 1)(1 1)(1 1)]^T or [(2 2)|(0 2)(1 1)(2 0)]^T
inline int check_T2_redundancy(const int m, const int K, const register int n_r[], const register int q[], const int num_P[]) {
	int i = 0, j = 0, l1 = 0, l2 = 0, l3 = 0;
	
	// check for [(2 2)|(1 1)(1 1)(1 1)]^T or [(2 2)|(0 2)(1 1)(2 0)]^T subconfig
	for (i = num_P[1]; i < (num_P[1] + num_P[2] - 1); i++) {
		for (j = i + 1; j < (num_P[1] + num_P[2]); j++) {
//			for (l1 = 0; l1 < (K - 2); l1++) {
//				for (l2 = l1 + 1; l2 < (K - 1); l2++) {
//					for (l3 = l2 + 1; l3 < K; l3++) {
			for (l1 = 0; l1 < K; l1++) {
				for (l2 = 0; l2 < K; l2++) {
					for (l3 = 0; l3 < K; l3++) {
						if ( (l1 != l2) && (l1 != l3) && (l2 != l3) ) {
							// [(2 2)|(1 1)(1 1)(1 1)]^T
							if ( (q[l1+(i*K)]==1) && (q[l1+(j*K)]==1) && (q[l2+(i*K)]==1) && (q[l2+(j*K)]==1) && (q[l3+(i*K)]==1) && (q[l3+(j*K)]==1) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==2) ) {
								return true;
							}
							// [(2 2)|(0 2)(1 1)(2 0)]^T
							if ( (q[l1+(i*K)]==0) && (q[l1+(j*K)]==2) && (q[l2+(i*K)]==1) && (q[l2+(j*K)]==1) && (q[l3+(i*K)]==2) && (q[l3+(j*K)]==0) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==2) ) {
								return true;
							}
						}
					}
				}
			}
		}
	}
	
	// check for [3|2 2] subconfig
	for (i = (num_P[1] + num_P[2]); i < (num_P[1] + num_P[2] + num_P[3]); i++) {
		for (l1 = 0; l1 < K; l1++) {
			for (l2 = 0; l2 < K; l2++) {
				if (l1 != l2) {
					// [3|2 2]
					if ( (q[l1+(i*K)]==2) && (q[l2+(i*K)]==2) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) ) {
						return true;
					}
				}
			}
		}
	}
	
	// EXTRA: NO T^2 AT ALL!!, SO check for [2|3] subconfig also (last remaining T^2 config)
	/*for (i = num_P[1]; i < (num_P[1] + num_P[2]); i++) {
		for (j=0; j < K; j++) {
			if ( (q[j+(i*K)]==3) && (sum_q_column(j,m,K,q)==3) ) {
				return true;
			}
		}
	}*/
	
	return false;
}

// returns TRUE if configuration matrix contains a K3 sub-configurations other than [3|4]
inline int check_K3_redundancy(const int m, const int K, const register int n_r[], const register int q[], const int num_P[]) {
	int i = 0, j = 0, l1 = 0, l2 = 0, l3 = 0, l4 = 0, l5 = 0, l6 = 0, i1 = 0, i2 = 0, i3 = 0, i4 = 0;
	
	// check for [4|2 3] subconfig
	/*for (i = (num_P[1] + num_P[2] + num_P[3]); i < (num_P[1] + num_P[2] + num_P[3] + num_P[4]); i++) {
		for (l1 = 0; l1 < K; l1++) {
			for (l2 = 0; l2 < K; l2++) {
				if (l1 != l2) {
					// [4|2 3]
					if ( (q[l1+(i*K)]==2) && (q[l2+(i*K)]==3) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==3) ) {
						return true;
					}
				}
			}
		}
	}*/
	
	// check for [5|2 2 2] subconfig
	/*for (i = (num_P[1] + num_P[2] + num_P[3] + num_P[4]); i < (num_P[1] + num_P[2] + num_P[3] + num_P[4] + num_P[5]); i++) {
		for (l1 = 0; l1 < K; l1++) {
			for (l2 = 0; l2 < K; l2++) {
				for (l3 = 0; l3 < K; l3++) {
					if ( (l1 != l2) && (l1 != l3) && (l2 != l3) ) {
						// [5|2 2 2]
						if ( (q[l1+(i*K)]==2) && (q[l2+(i*K)]==2) && (q[l3+(i*K)]==2) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==2) ) {
							return true;
						}
					}
				}
			}
		}
	}*/
	
	// check for [(2 2)|(1 1)(2 2)]^T or [(2 2)|(0 2)(3 1)]^T or [(2 2)|(1 2)(2 1)]^T subconfig
	/*for (i = num_P[1]; i < (num_P[1] + num_P[2] - 1); i++) {
		for (j = i + 1; j < (num_P[1] + num_P[2]); j++) {
			for (l1 = 0; l1 < K; l1++) {
				for (l2 = 0; l2 < K; l2++) {
					if (l1 != l2) {
						// [(2 2)|(1 1)(2 2)]^T
						if ( (q[l1+(i*K)]==1) && (q[l1+(j*K)]==1) && (q[l2+(i*K)]==2) && (q[l2+(j*K)]==2) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==4) ) {
							return true;
						}
						// [(2 2)|(0 2)(3 1)]^T
						if ( (q[l1+(i*K)]==0) && (q[l1+(j*K)]==2) && (q[l2+(i*K)]==3) && (q[l2+(j*K)]==1) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==4) ) {
							return true;
						}
						// [(2 2)|(1 2)(2 1)]^T
						if ( (q[l1+(i*K)]==1) && (q[l1+(j*K)]==2) && (q[l2+(i*K)]==2) && (q[l2+(j*K)]==1) && (sum_q_column(l1,m,K,q)==3) && (sum_q_column(l2,m,K,q)==3) ) {
							return true;
						}
					}
				}
			}
		}
	}*/
	
	// check for [(2 3)|...]^T subconfigs
	for (i = num_P[1]; i < (num_P[1] + num_P[2]); i++) {
		for (j = (num_P[1] + num_P[2]); j < (num_P[1] + num_P[2] + num_P[3]); j++) {
			for (l1 = 0; l1 < K; l1++) {
				for (l2 = 0; l2 < K; l2++) {
					for (l3 = 0; l3 < K; l3++) {
						if ( (l1 != l2) && (l1 != l3) && (l2 != l3) ) {
							// [(2 3)|(1 1)(1 1)(1 2)]^T
							if ( (q[l1+(i*K)]==1) && (q[l1+(j*K)]==1) && (q[l2+(i*K)]==1) && (q[l2+(j*K)]==1) && (q[l3+(i*K)]==1) && (q[l3+(j*K)]==2) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==3) ) {
								return true;
							}
							// [(2 3)|(0 2)(1 1)(2 1)]^T
							if ( (q[l1+(i*K)]==0) && (q[l1+(j*K)]==2) && (q[l2+(i*K)]==1) && (q[l2+(j*K)]==1) && (q[l3+(i*K)]==2) && (q[l3+(j*K)]==1) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==3) ) {
								return true;
							}
							// [(2 3)|(0 2)(1 2)(2 0)]^T
							if ( (q[l1+(i*K)]==0) && (q[l1+(j*K)]==2) && (q[l2+(i*K)]==1) && (q[l2+(j*K)]==2) && (q[l3+(i*K)]==2) && (q[l3+(j*K)]==0) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==3) && (sum_q_column(l3,m,K,q)==2) ) {
								return true;
							}
							// [(2 3)|(0 3)(1 1)(2 0)]^T
							if ( (q[l1+(i*K)]==0) && (q[l1+(j*K)]==3) && (q[l2+(i*K)]==1) && (q[l2+(j*K)]==1) && (q[l3+(i*K)]==2) && (q[l3+(j*K)]==0) && (sum_q_column(l1,m,K,q)==3) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==2) ) {
								return true;
							}							
						}
					}
				}
			}
		}
	}

	// check for [(2 4)|...]^T subconfigs
	for (i = num_P[1]; i < (num_P[1] + num_P[2]); i++) {
		for (j = (num_P[1] + num_P[2] + num_P[3]); j < (num_P[1] + num_P[2] + num_P[3] + num_P[4]); j++) {
			for (l1 = 0; l1 < K; l1++) {
				for (l2 = 0; l2 < K; l2++) {
					for (l3 = 0; l3 < K; l3++) {
						for (l4 = 0; l4 < K; l4++) {
							if ( (l1 != l2) && (l1 != l3) && (l1 != l4) && (l2 != l3) && (l2 != l4) && (l3 != l4) ) {
								// [(2 4)|(0 2)(1 1)(1 1)(1 1)]^T
								if ( (q[l1+(i*K)]==0) && (q[l1+(j*K)]==2) && (q[l2+(i*K)]==1) && (q[l2+(j*K)]==1) && (q[l3+(i*K)]==1) && (q[l3+(j*K)]==1) && (q[l4+(i*K)]==1) && (q[l4+(j*K)]==1) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==2) && (sum_q_column(l4,m,K,q)==2) ) {
									return true;
								}
								// [(2 4)|(0 2)(0 2)(1 1)(2 0)]^T
								if ( (q[l1+(i*K)]==0) && (q[l1+(j*K)]==2) && (q[l2+(i*K)]==0) && (q[l2+(j*K)]==2) && (q[l3+(i*K)]==1) && (q[l3+(j*K)]==1) && (q[l4+(i*K)]==2) && (q[l4+(j*K)]==0) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==2) && (sum_q_column(l4,m,K,q)==2) ) {
									return true;
								}								
							}
						}
					}
				}
			}
		}
	}
	
	// check for [(3 3)|...]^T subconfigs
	for (i = (num_P[1] + num_P[2]); i < (num_P[1] + num_P[2] + num_P[3] - 1); i++) {
		for (j = i + 1; j < (num_P[1] + num_P[2] + num_P[3]); j++) {
			for (l1 = 0; l1 < K; l1++) {
				for (l2 = 0; l2 < K; l2++) {
					for (l3 = 0; l3 < K; l3++) {
						for (l4 = 0; l4 < K; l4++) {
							if ( (l1 != l2) && (l1 != l3) && (l1 != l4) && (l2 != l3) && (l2 != l4) && (l3 != l4) ) {
								// [(3 3)|(1 1)(1 1)(1 1)(1 1)]^T
								if ( (q[l1+(i*K)]==1) && (q[l1+(j*K)]==1) && (q[l2+(i*K)]==1) && (q[l2+(j*K)]==1) && (q[l3+(i*K)]==1) && (q[l3+(j*K)]==1) && (q[l4+(i*K)]==1) && (q[l4+(j*K)]==1) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==2) && (sum_q_column(l4,m,K,q)==2) ) {
									return true;
								}
								// [(3 3)|(0 2)(1 1)(1 1)(2 0)]^T
								if ( (q[l1+(i*K)]==0) && (q[l1+(j*K)]==2) && (q[l2+(i*K)]==1) && (q[l2+(j*K)]==1) && (q[l3+(i*K)]==1) && (q[l3+(j*K)]==1) && (q[l4+(i*K)]==2) && (q[l4+(j*K)]==0) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==2) && (sum_q_column(l4,m,K,q)==2) ) {
									return true;
								}								
							}
						}
					}
				}
			}
		}
	}
	
	// check for [(2 2 2)|...]^T subconfigs
	for (i1 = num_P[1]; i1 < (num_P[1] + num_P[2] - 2); i1++) {
		for (i2 = i1 + 1; i2 < (num_P[1] + num_P[2] - 1); i2++) {
			for (i3 = i2 + 1; i3 < (num_P[1] + num_P[2]); i3++) {
				for (l1 = 0; l1 < K; l1++) {
					for (l2 = 0; l2 < K; l2++) {
						for (l3 = 0; l3 < K; l3++) {
							for (l4 = 0; l4 < K; l4++) {
								if ( (l1 != l2) && (l1 != l3) && (l1 != l4) && (l2 != l3) && (l2 != l4) && (l3 != l4) ) {
									// [(2 2 2)|(0 1 1)(1 0 1)(1 1 0)(1 1 1)]^T
									if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==1) && (q[l1+(i3*K)]==1) && (q[l2+(i1*K)]==1) && (q[l2+(i2*K)]==0) && (q[l2+(i3*K)]==1) && (q[l3+(i1*K)]==1) && (q[l3+(i2*K)]==1) && (q[l3+(i3*K)]==0) && (q[l4+(i1*K)]==1) && (q[l4+(i2*K)]==1) && (q[l4+(i3*K)]==1) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==2) && (sum_q_column(l4,m,K,q)==3) ) {
										return true;
									}
									// [(2 2 2)|(0 1 1)(1 0 1)(1 0 1)(1 2 0)]^T
									if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==1) && (q[l1+(i3*K)]==1) && (q[l2+(i1*K)]==1) && (q[l2+(i2*K)]==0) && (q[l2+(i3*K)]==1) && (q[l3+(i1*K)]==1) && (q[l3+(i2*K)]==0) && (q[l3+(i3*K)]==1) && (q[l4+(i1*K)]==1) && (q[l4+(i2*K)]==2) && (q[l4+(i3*K)]==0) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==2) && (sum_q_column(l4,m,K,q)==3) ) {
										return true;
									}
									// [(2 2 2)|(0 1 1)(0 1 1)(1 1 1)(2 0 0)]^T
									if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==1) && (q[l1+(i3*K)]==1) && (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==1) && (q[l2+(i3*K)]==1) && (q[l3+(i1*K)]==1) && (q[l3+(i2*K)]==1) && (q[l3+(i3*K)]==1) && (q[l4+(i1*K)]==2) && (q[l4+(i2*K)]==0) && (q[l4+(i3*K)]==0) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==3) && (sum_q_column(l4,m,K,q)==2) ) {
										return true;
									}
									// [(2 2 2)|(0 1 1)(0 1 2)(1 1 0)(2 0 0)]^T
									if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==1) && (q[l1+(i3*K)]==1) && (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==1) && (q[l2+(i3*K)]==2) && (q[l3+(i1*K)]==1) && (q[l3+(i2*K)]==1) && (q[l3+(i3*K)]==0) && (q[l4+(i1*K)]==2) && (q[l4+(i2*K)]==0) && (q[l4+(i3*K)]==0) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==3) && (sum_q_column(l3,m,K,q)==2) && (sum_q_column(l4,m,K,q)==2) ) {
										return true;
									}
									// [(2 2 2)|(0 0 2)(0 2 0)(1 1 0)(2 0 1)]^T
									if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==0) && (q[l1+(i3*K)]==2) && (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==2) && (q[l2+(i3*K)]==0) && (q[l3+(i1*K)]==1) && (q[l3+(i2*K)]==1) && (q[l3+(i3*K)]==0) && (q[l4+(i1*K)]==2) && (q[l4+(i2*K)]==0) && (q[l4+(i3*K)]==1) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==2) && (sum_q_column(l4,m,K,q)==3) ) {
										return true;
									}
									// [(2 2 2)|(0 0 2)(0 2 0)(1 1 1)(2 0 0)]^T
									if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==0) && (q[l1+(i3*K)]==2) && (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==2) && (q[l2+(i3*K)]==0) && (q[l3+(i1*K)]==1) && (q[l3+(i2*K)]==1) && (q[l3+(i3*K)]==1) && (q[l4+(i1*K)]==2) && (q[l4+(i2*K)]==0) && (q[l4+(i3*K)]==0) && (sum_q_column(l1,m,K,q)==2) && (sum_q_column(l2,m,K,q)==2) && (sum_q_column(l3,m,K,q)==3) && (sum_q_column(l4,m,K,q)==2) ) {
										return true;
									}									
								}
							}
						}
					}
				}
			}
		}
	}
	
	// check for [(2 2 3)|...]^T subconfigs
	for (i1 = num_P[1]; i1 < (num_P[1] + num_P[2] - 1); i1++) {
		for (i2 = i1 + 1; i2 < (num_P[1] + num_P[2]); i2++) {
			for (i3 = (num_P[1] + num_P[2]); i3 < (num_P[1] + num_P[2] + num_P[3]); i3++) {
				for (l1 = 0; l1 < K; l1++) {
					for (l2 = 0; l2 < K; l2++) {
						for (l3 = 0; l3 < K; l3++) {
							for (l4 = 0; l4 < K; l4++) {
								for (l5 = 0; l5 < K; l5++) {
									if ( (l1 != l2) && (l1 != l3) && (l1 != l4) && (l1 != l5) && (l2 != l3) && (l2 != l4) && (l2 != l5) && (l3 != l4) && (l3 != l5) && (l4 != l5) ) {
										// [(2 2 3)|(0 1 1)(0 2 0)(1 0 1)(1 0 1)(1 0 1)]^T
										if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==1) && (q[l1+(i3*K)]==1) && 
											 (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==2) && (q[l2+(i3*K)]==0) && 
											 (q[l3+(i1*K)]==1) && (q[l3+(i2*K)]==0) && (q[l3+(i3*K)]==1) && 
											 (q[l4+(i1*K)]==1) && (q[l4+(i2*K)]==0) && (q[l4+(i3*K)]==1) && 
											 (q[l5+(i1*K)]==1) && (q[l5+(i2*K)]==0) && (q[l5+(i3*K)]==1) &&
											 (sum_q_column(l1,m,K,q)==2) && 
											 (sum_q_column(l2,m,K,q)==2) && 
											 (sum_q_column(l3,m,K,q)==2) && 
											 (sum_q_column(l4,m,K,q)==2) &&
											 (sum_q_column(l5,m,K,q)==2) ) {
											return true;
										}
										// [(2 2 3)|(0 1 1)(0 1 1)(1 0 1)(1 0 1)(1 1 0)]^T
										if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==1) && (q[l1+(i3*K)]==1) && 
											 (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==1) && (q[l2+(i3*K)]==1) && 
											 (q[l3+(i1*K)]==1) && (q[l3+(i2*K)]==0) && (q[l3+(i3*K)]==1) && 
											 (q[l4+(i1*K)]==1) && (q[l4+(i2*K)]==0) && (q[l4+(i3*K)]==1) && 
											 (q[l5+(i1*K)]==1) && (q[l5+(i2*K)]==1) && (q[l5+(i3*K)]==0) &&
											 (sum_q_column(l1,m,K,q)==2) && 
											 (sum_q_column(l2,m,K,q)==2) && 
											 (sum_q_column(l3,m,K,q)==2) && 
											 (sum_q_column(l4,m,K,q)==2) &&
											 (sum_q_column(l5,m,K,q)==2) ) {
											return true;
										}
										// [(2 2 3)|(0 0 2)(0 2 0)(1 0 1)(1 0 1)(1 1 0)]^T
										if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==0) && (q[l1+(i3*K)]==2) && 
											 (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==2) && (q[l2+(i3*K)]==0) && 
											 (q[l3+(i1*K)]==1) && (q[l3+(i2*K)]==0) && (q[l3+(i3*K)]==1) && 
											 (q[l4+(i1*K)]==1) && (q[l4+(i2*K)]==0) && (q[l4+(i3*K)]==1) && 
											 (q[l5+(i1*K)]==1) && (q[l5+(i2*K)]==1) && (q[l5+(i3*K)]==0) &&
											 (sum_q_column(l1,m,K,q)==2) && 
											 (sum_q_column(l2,m,K,q)==2) && 
											 (sum_q_column(l3,m,K,q)==2) && 
											 (sum_q_column(l4,m,K,q)==2) &&
											 (sum_q_column(l5,m,K,q)==2) ) {
											return true;
										}
										// [(2 2 3)|(0 0 2)(0 1 1)(0 2 0)(1 0 1)(2 0 0)]^T
										if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==0) && (q[l1+(i3*K)]==2) && 
											 (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==1) && (q[l2+(i3*K)]==1) && 
											 (q[l3+(i1*K)]==0) && (q[l3+(i2*K)]==2) && (q[l3+(i3*K)]==0) && 
											 (q[l4+(i1*K)]==1) && (q[l4+(i2*K)]==0) && (q[l4+(i3*K)]==1) && 
											 (q[l5+(i1*K)]==2) && (q[l5+(i2*K)]==0) && (q[l5+(i3*K)]==0) &&
											 (sum_q_column(l1,m,K,q)==2) && 
											 (sum_q_column(l2,m,K,q)==2) && 
											 (sum_q_column(l3,m,K,q)==2) && 
											 (sum_q_column(l4,m,K,q)==2) &&
											 (sum_q_column(l5,m,K,q)==2) ) {
											return true;
										}
										// [(2 2 3)|(0 0 2)(0 1 1)(1 0 1)(1 1 0)(1 1 0)]^T
										if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==0) && (q[l1+(i3*K)]==2) && 
											 (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==1) && (q[l2+(i3*K)]==1) && 
											 (q[l3+(i1*K)]==1) && (q[l3+(i2*K)]==0) && (q[l3+(i3*K)]==1) && 
											 (q[l4+(i1*K)]==1) && (q[l4+(i2*K)]==1) && (q[l4+(i3*K)]==0) && 
											 (q[l5+(i1*K)]==1) && (q[l5+(i2*K)]==1) && (q[l5+(i3*K)]==0) &&
											 (sum_q_column(l1,m,K,q)==2) && 
											 (sum_q_column(l2,m,K,q)==2) && 
											 (sum_q_column(l3,m,K,q)==2) && 
											 (sum_q_column(l4,m,K,q)==2) &&
											 (sum_q_column(l5,m,K,q)==2) ) {
											return true;
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
	
	// check for [(2 2 2 2)|...]^T subconfigs
	for (i1 = num_P[1]; i1 < (num_P[1] + num_P[2] - 3); i1++) {
		for (i2 = i1 + 1; i2 < (num_P[1] + num_P[2] - 2); i2++) {
			for (i3 = i2 + 1; i3 < (num_P[1] + num_P[2] - 1); i3++) {
				for (i4 = i3 + 1; i4 < (num_P[1] + num_P[2]); i4++) {
					for (l1 = 0; l1 < K; l1++) {
						for (l2 = 0; l2 < K; l2++) {
							for (l3 = 0; l3 < K; l3++) {
								for (l4 = 0; l4 < K; l4++) {
									for (l5 = 0; l5 < K; l5++) {
										for (l6 = 0; l6 < K; l6++) {
											if ( (l1 != l2) && (l1 != l3) && (l1 != l4) && (l1 != l5) && (l1 != l6) && (l2 != l3) && (l2 != l4) && (l2 != l5) && (l2 != l6) && (l3 != l4) && (l3 != l5) && (l3 != l6) && (l4 != l5) && (l4 != l6) && (l5 != l6) ) {
												// [(2 2 2 2)|(0 0 1 1)(0 1 0 1)(0 1 1 0)(0 1 1 0)(1 0 0 1)(2 0 0 0)]^T
												if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==0) && (q[l1+(i3*K)]==1) && (q[l1+(i4*K)]==1) && 
													 (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==1) && (q[l2+(i3*K)]==0) && (q[l2+(i4*K)]==1) &&
													 (q[l3+(i1*K)]==0) && (q[l3+(i2*K)]==1) && (q[l3+(i3*K)]==1) && (q[l3+(i4*K)]==0) &&
													 (q[l4+(i1*K)]==0) && (q[l4+(i2*K)]==1) && (q[l4+(i3*K)]==1) && (q[l4+(i4*K)]==0) &&
													 (q[l5+(i1*K)]==1) && (q[l5+(i2*K)]==0) && (q[l5+(i3*K)]==0) && (q[l5+(i4*K)]==1) &&
													 (q[l6+(i1*K)]==2) && (q[l6+(i2*K)]==0) && (q[l6+(i3*K)]==0) && (q[l6+(i4*K)]==0) &&
													 (sum_q_column(l1,m,K,q)==2) && 
													 (sum_q_column(l2,m,K,q)==2) && 
													 (sum_q_column(l3,m,K,q)==2) && 
													 (sum_q_column(l4,m,K,q)==2) &&
													 (sum_q_column(l5,m,K,q)==2) &&
													 (sum_q_column(l6,m,K,q)==2) ) {
													return true;
												}
												// [(2 2 2 2)|(0 0 1 1)(0 1 0 1)(0 1 1 0)(1 0 0 1)(1 0 1 0)(1 1 0 0)]^T
												if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==0) && (q[l1+(i3*K)]==1) && (q[l1+(i4*K)]==1) && 
													 (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==1) && (q[l2+(i3*K)]==0) && (q[l2+(i4*K)]==1) &&
													 (q[l3+(i1*K)]==0) && (q[l3+(i2*K)]==1) && (q[l3+(i3*K)]==1) && (q[l3+(i4*K)]==0) &&
													 (q[l4+(i1*K)]==1) && (q[l4+(i2*K)]==0) && (q[l4+(i3*K)]==0) && (q[l4+(i4*K)]==1) &&
													 (q[l5+(i1*K)]==1) && (q[l5+(i2*K)]==0) && (q[l5+(i3*K)]==1) && (q[l5+(i4*K)]==0) &&
													 (q[l6+(i1*K)]==1) && (q[l6+(i2*K)]==1) && (q[l6+(i3*K)]==0) && (q[l6+(i4*K)]==0) &&
													 (sum_q_column(l1,m,K,q)==2) && 
													 (sum_q_column(l2,m,K,q)==2) && 
													 (sum_q_column(l3,m,K,q)==2) && 
													 (sum_q_column(l4,m,K,q)==2) &&
													 (sum_q_column(l5,m,K,q)==2) &&
													 (sum_q_column(l6,m,K,q)==2) ) {
													 return true;
												}
												// [(2 2 2 2)|(0 0 1 1)(0 1 0 1)(0 1 0 1)(1 0 1 0)(1 0 1 0)(1 1 0 0)]^T
												if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==0) && (q[l1+(i3*K)]==1) && (q[l1+(i4*K)]==1) && 
													 (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==1) && (q[l2+(i3*K)]==0) && (q[l2+(i4*K)]==1) &&
													 (q[l3+(i1*K)]==0) && (q[l3+(i2*K)]==1) && (q[l3+(i3*K)]==0) && (q[l3+(i4*K)]==1) &&
													 (q[l4+(i1*K)]==1) && (q[l4+(i2*K)]==0) && (q[l4+(i3*K)]==1) && (q[l4+(i4*K)]==0) &&
													 (q[l5+(i1*K)]==1) && (q[l5+(i2*K)]==0) && (q[l5+(i3*K)]==1) && (q[l5+(i4*K)]==0) &&
													 (q[l6+(i1*K)]==1) && (q[l6+(i2*K)]==1) && (q[l6+(i3*K)]==0) && (q[l6+(i4*K)]==0) &&
													 (sum_q_column(l1,m,K,q)==2) && 
													 (sum_q_column(l2,m,K,q)==2) && 
													 (sum_q_column(l3,m,K,q)==2) && 
													 (sum_q_column(l4,m,K,q)==2) &&
													 (sum_q_column(l5,m,K,q)==2) &&
													 (sum_q_column(l6,m,K,q)==2) ) {
													return true;
												}
												// [(2 2 2 2)|(0 0 1 1)(0 0 1 1)(0 1 0 1)(0 2 0 0)(1 0 1 0)(2 0 0 0)]^T
												if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==0) && (q[l1+(i3*K)]==1) && (q[l1+(i4*K)]==1) && 
													 (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==0) && (q[l2+(i3*K)]==1) && (q[l2+(i4*K)]==1) &&
													 (q[l3+(i1*K)]==0) && (q[l3+(i2*K)]==1) && (q[l3+(i3*K)]==0) && (q[l3+(i4*K)]==1) &&
													 (q[l4+(i1*K)]==0) && (q[l4+(i2*K)]==2) && (q[l4+(i3*K)]==0) && (q[l4+(i4*K)]==0) &&
													 (q[l5+(i1*K)]==1) && (q[l5+(i2*K)]==0) && (q[l5+(i3*K)]==1) && (q[l5+(i4*K)]==0) &&
													 (q[l6+(i1*K)]==2) && (q[l6+(i2*K)]==0) && (q[l6+(i3*K)]==0) && (q[l6+(i4*K)]==0) &&
													 (sum_q_column(l1,m,K,q)==2) && 
													 (sum_q_column(l2,m,K,q)==2) && 
													 (sum_q_column(l3,m,K,q)==2) && 
													 (sum_q_column(l4,m,K,q)==2) &&
													 (sum_q_column(l5,m,K,q)==2) &&
													 (sum_q_column(l6,m,K,q)==2) ) {
													return true;
												}
												// [(2 2 2 2)|(0 0 0 2)(0 0 2 0)(0 2 0 0)(1 0 0 1)(1 0 1 0)(1 1 0 0)]^T
												if ( (q[l1+(i1*K)]==0) && (q[l1+(i2*K)]==0) && (q[l1+(i3*K)]==0) && (q[l1+(i4*K)]==2) && 
													 (q[l2+(i1*K)]==0) && (q[l2+(i2*K)]==0) && (q[l2+(i3*K)]==2) && (q[l2+(i4*K)]==0) &&
													 (q[l3+(i1*K)]==0) && (q[l3+(i2*K)]==2) && (q[l3+(i3*K)]==0) && (q[l3+(i4*K)]==0) &&
													 (q[l4+(i1*K)]==1) && (q[l4+(i2*K)]==0) && (q[l4+(i3*K)]==0) && (q[l4+(i4*K)]==1) &&
													 (q[l5+(i1*K)]==1) && (q[l5+(i2*K)]==0) && (q[l5+(i3*K)]==1) && (q[l5+(i4*K)]==0) &&
													 (q[l6+(i1*K)]==1) && (q[l6+(i2*K)]==1) && (q[l6+(i3*K)]==0) && (q[l6+(i4*K)]==0) &&
													 (sum_q_column(l1,m,K,q)==2) && 
													 (sum_q_column(l2,m,K,q)==2) && 
													 (sum_q_column(l3,m,K,q)==2) && 
													 (sum_q_column(l4,m,K,q)==2) &&
													 (sum_q_column(l5,m,K,q)==2) &&
													 (sum_q_column(l6,m,K,q)==2) ) {
													return true;
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}

	return false;
}

// returns true if the row is contractible or false otherwise
inline int is_contractible_row(const int row, const int K, const register int q[]) {
	int i;

	for (i = 0; i < K; i++) {
		if (q[i+(row*K)] > 1) {
			return false;
		}
	}

	return true;
}

// returns smallest possible contraction (opposite of ineffective splitting) of q-matrix or NULL if it cannot be contracted further
inline conf* contract(const int old_m, const int old_K, const int old_n_r[], const int old_q[], const int old_eulernumber) {
	int i, j, l, l1, l2, l3;
	int new_m, new_K, next_free_new_col;
	int * new_n_r, * new_q;
	int new_eulernumber;
	conf* new_conf;
	
	int ** u;
	int tmp_K, nu, ineff;
	
	for (i = 0; i < old_m; i++) {
		// for n>3, only consider P1 splits/contractions for the moment:
		//if ( (n > 3) && (old_n_r[i] > 1) ) {break;}
	
		if (is_contractible_row(i, old_K, old_q) == true) {
			// now, do the contraction:
			
			// 2nd reserve space for contracted matrix
			new_m = (old_m-1);
			new_n_r = init_array(new_m, 0);
			for (j = 0; j < i; j++) {
				new_n_r[j] = old_n_r[j];
			}
			for (j = i+1; j < old_m; j++) {
				new_n_r[j-1] = old_n_r[j];
			}
			new_K = old_K - old_n_r[i];
			new_q = init_array(new_m*new_K, 0);
			next_free_new_col = 1;
			for (j = 0; j < old_K; j++) {
				if (old_q[j+(i*old_K)] == 0) {
					// just copy the whole (m-1) column vector
					for (l = 0; l < i; l++) {
						new_q[next_free_new_col+(l*new_K)] = old_q[j+(l*old_K)];
					}
					for (l = i+1; l < old_m; l++) {
						new_q[next_free_new_col+((l-1)*new_K)] = old_q[j+(l*old_K)];
					}
					next_free_new_col++;
				}
				if (old_q[j+(i*old_K)] == 1) {
					// add this column to first column of new q-matrix
					for (l = 0; l < i; l++) {
						new_q[(l*new_K)] += old_q[j+(l*old_K)];
					}
					for (l = i+1; l < old_m; l++) {
						new_q[((l-1)*new_K)] += old_q[j+(l*old_K)];
					}
				}
			}
			
			// start: implement (3.24)-(3.25) of Candelas et. al. orig CICY 3-fold paper (NBP 298 (1988) 493-525)
			// 1) initialize u_1 ... u_{n+1}
			nu = 0;
			u = new int*[old_n_r[i] + 1];
			l = 0;
			for (j = 0; j < old_K; j++) {
				if (old_q[j + (i*old_K)] == 1) {
					u[l] = new int[old_m - 1];
					for (l1 = 0; l1 < old_m; l1++) {
						if (l1 != i) {
							u[l][l1 - ((l1 > i) ? 1 : 0)] = old_q[j + (l1*old_K)];
						}
					}
					l++;
				}
			}
			//debug_output(old_m, old_K, old_n_r, old_q); printf("i=%i\n", i+1);
			tmp_K = old_K - (old_n_r[i] + 1) + 4;
			if (n==3) {
				// 2a) process first term in (3.25)
				for (j = 0; j < (old_n_r[i] + 1); j++) {
					for (l = j+1; l < (old_n_r[i] + 1); l++) {
						nu += build_and_calc_subdetvar(3, old_m, old_K, old_q, tmp_K, i, new_n_r, u[j], u[j], u[l], u[l], NULL);
					}
				}
				// 2b) process second term in (3.25)
				if ( old_n_r[i] > 1 ) {
					for (j = 0; j < (old_n_r[i] + 1); j++) {
						for (l = j+1; l < (old_n_r[i] + 1); l++) {
							for (l1 = 0; l1 < (old_n_r[i] + 1); l1++) {
								if ( (l1 != j) && (l1 != l) ) {
									//printf("i=%i, (%i, %i, %i, %i)\n", i+1, j+1, l+1, l1+1, l1+1);
									nu += build_and_calc_subdetvar(3, old_m, old_K, old_q, tmp_K, i, new_n_r, u[j], u[l], u[l1], u[l1], NULL);
								}
							}
						}
					}
				}
				// 2c) process third term in (3.25)
				if ( old_n_r[i] > 2 ) {
					for (j = 0; j < (old_n_r[i] + 1); j++) {
						for (l = j+1; l < (old_n_r[i] + 1); l++) {
							for (l1 = l+1; l1 < (old_n_r[i] + 1); l1++) {
								for (l2 = l1+1; l2 < (old_n_r[i] + 1); l2++) {
									//printf("i=%i, (%i, %i, %i, %i)\n", i+1, j+1, l+1, l1+1, l2+1);
									nu += 2 * build_and_calc_subdetvar(3, old_m, old_K, old_q, tmp_K, i, new_n_r, u[j], u[l], u[l1], u[l2], NULL);
								}
							}
						}
					}
				}
				ineff = (nu == 0) ? true : false;
			}
			else if (n==4) {
				if (old_n_r[i] == 1) {
					int empty;
					nu = build_and_calc_subdetvar(4, old_m, old_K, old_q, tmp_K, i, new_n_r, u[0], u[0], u[1], u[1], &empty);
					if ( (nu == 0) && (empty == true) ) {
						ineff = true;
					}
					else {
						ineff = false;
					}
				}
				else {
					// 1st: compute nu using (2.13) of NBP 498 (1997) 156-174
					// a) process first term in (2.13)
					for (j = 0; j < (old_n_r[i] + 1); j++) {
						for (l = 0; l < (old_n_r[i] + 1); l++) {
							for (l1 = l+1; l1 < (old_n_r[i] + 1); l1++) {
								//printf("i=%i, 1, (%i, %i, %i, %i, %i)\n", i+1, j+1, j+1, j+1, l+1, l1+1);
								nu += 4 * calc_nu_comp(old_m, old_K, old_q, tmp_K + 1, i, new_n_r, u[j], u[j], u[j], u[l], u[l1]);
							}
						}
					}
					// b) process second term in (2.13)
					for (j = 0; j < (old_n_r[i] + 1); j++) {
						for (l = 0; l < (old_n_r[i] + 1); l++) {
							for (l1 = l+1; l1 < (old_n_r[i] + 1); l1++) {
								for (l2 = l1+1; l2 < (old_n_r[i] + 1); l2++) {
									//printf("i=%i, 2, (%i, %i, %i, %i, %i)\n", i+1, j+1, j+1, l+1, l1+1, l2+1);
									nu += 6 * calc_nu_comp(old_m, old_K, old_q, tmp_K + 1, i, new_n_r, u[j], u[j], u[l], u[l1], u[l2]);
								}
							}
						}
					}
					// c) process third term in (2.13)
					for (j = 0; j < (old_n_r[i] + 1); j++) {
						for (l = 0; l < (old_n_r[i] + 1); l++) {
							for (l1 = l+1; l1 < (old_n_r[i] + 1); l1++) {
								for (l2 = l1+1; l2 < (old_n_r[i] + 1); l2++) {
									for (l3 = l2+1; l3 < (old_n_r[i] + 1); l3++) {
										//printf("i=%i, 3, (%i, %i, %i, %i, %i)\n", i+1, j+1, l+1, l1+1, l2+1, l3+1);
										nu += 4 * calc_nu_comp(old_m, old_K, old_q, tmp_K + 1, i, new_n_r, u[j], u[l], u[l1], u[l2], u[l3]);
									}
								}
							}
						}
					}
					// d) process fourth term in (2.13)
					for (j = 0; j < (old_n_r[i] + 1); j++) {
						for (l = 0; l < (old_n_r[i] + 1); l++) {
							for (l1 = 0; l1 < (old_n_r[i] + 1); l1++) {
								for (l2 = 0; l2 < (old_n_r[i] + 1); l2++) {
									for (l3 = 0; l3 < (old_n_r[i] + 1); l3++) {
										if (!( (l == l1) && (l1 == l2) && (l2 == l3) )) {
											//printf("i=%i, 4, (%i, %i, %i, %i, %i)\n", i+1, j+1, l+1, l1+1, l2+1, l3+1);
											nu -= calc_nu_comp(old_m, old_K, old_q, tmp_K + 1, i, new_n_r, u[j], u[l], u[l1], u[l2], u[l3]);
										}
									}
								}
							}
						}
					}
					nu /= 10; // coefficient in front of (2.13); NOTE: no *3 needed as this is taken care of in the general expression "new_eulernumber = ..." below
					// debug:
					//int tmp_eulernumber = compute_eulernumber(new_m, new_K, new_n_r, new_q);
					//if ( tmp_eulernumber != (old_eulernumber - (n-1)*nu) ) {
					//	printf("Pn=%i, nu=%i, orig_euler=%i, newcode_euler=%i, full_euler=%i\n", old_n_r[i], nu, old_eulernumber, old_eulernumber - (n-1)*nu, tmp_eulernumber); // debug
					//	debug_output(old_m, old_K, old_n_r, old_q);
					//}
					
					if (nu != 0) { // euler number changing effective splitting
						ineff = false;
					}
					else {
						ineff = true; 
						//debug_output(old_m, old_K, old_n_r, old_q); printf("i=%i\n", i); // debug
						// 2nd: intersect w/ hyperplanes defined by unit column vectors (see paper ITP-UH-04/13)
						int extra_K = old_K + 1;
						int extra_q[old_m * extra_K];
						//int tmp_nu;
						for (j = 0; j < old_m; j++) {
							for (l = 0; l < extra_K; l++) {
								if (l == (extra_K-1)) {
									extra_q[l + (j*extra_K)] = 0;
								}
								else {
									extra_q[l + (j*extra_K)] = old_q[l + (j*old_K)];
								}
							}
						}
						for (j = old_m - 1; j >= 0 ; j--) {
							extra_q[(extra_K - 1) + (j*extra_K)] = 1;
							if (j != (old_m - 1)) {
								extra_q[(extra_K - 1) + ((j+1)*extra_K)] = 0;
							}
							if (j == i) {continue;}
							//debug_output(old_m, extra_K, old_n_r, extra_q);
							// now need to check if singular locus of the non-CY 3-fold split is empty or not
							//tmp_nu = 0;
							// 2a) process first term in (3.25)
							for (l = 0; l < (old_n_r[i] + 1); l++) {
								for (l1 = l+1; l1 < (old_n_r[i] + 1); l1++) {
									//tmp_nu += build_and_calc_subdetvar(3, old_m, extra_K, extra_q, tmp_K + 1, i, new_n_r, u[l], u[l], u[l1], u[l1], NULL);
									if (build_and_calc_subdetvar(3, old_m, extra_K, extra_q, tmp_K + 1, i, new_n_r, u[l], u[l], u[l1], u[l1], NULL) != 0) {
									// ^^ i.e. the intersection of the "unit column vector hyperplane" with the singular locus is non-zero 
									// ==> singular locus is not the empty set
									// ==> euler number preserving effective splitting
									// (tmp_nu is additive, so one non-zero term is enough to conclude it's non-zero. Don't need to continue computing full tmp_nu. Instead break out of calculation at first non-zero term.)
										ineff = false;
										break;
									}
								}
								if (ineff == false) {break;}
							}
							if (ineff == false) {break;}
							// 2b) process second term in (3.25)
							if ( old_n_r[i] > 1 ) {
								for (l = 0; l < (old_n_r[i] + 1); l++) {
									for (l1 = l+1; l1 < (old_n_r[i] + 1); l1++) {
										for (l2 = 0; l2 < (old_n_r[i] + 1); l2++) {
											if ( (l2 != l) && (l2 != l1) ) {
												//printf("i=%i, (%i, %i, %i, %i)\n", i+1, j+1, l+1, l1+1, l1+1);
												//tmp_nu += build_and_calc_subdetvar(3, old_m, extra_K, extra_q, tmp_K + 1, i, new_n_r, u[l], u[l1], u[l2], u[l2], NULL);
												if (build_and_calc_subdetvar(3, old_m, extra_K, extra_q, tmp_K + 1, i, new_n_r, u[l], u[l1], u[l2], u[l2], NULL) != 0) {
												// ^^ i.e. the intersection of the "unit column vector hyperplane" with the singular locus is non-zero 
												// ==> singular locus is not the empty set
												// ==> euler number preserving effective splitting
												// (tmp_nu is additive, so one non-zero term is enough to conclude it's non-zero. Don't need to continue computing full tmp_nu. Instead break out of calculation at first non-zero term.)
													ineff = false;
													break;
												}
											}
										}
										if (ineff == false) {break;}
									}
									if (ineff == false) {break;}
								}
							}
							if (ineff == false) {break;}
							// 2c) process third term in (3.25)
							if ( old_n_r[i] > 2 ) {
								for (l = 0; l < (old_n_r[i] + 1); l++) {
									for (l1 = l+1; l1 < (old_n_r[i] + 1); l1++) {
										for (l2 = l1+1; l2 < (old_n_r[i] + 1); l2++) {
											for (l3 = l2+1; l3 < (old_n_r[i] + 1); l3++) {
												//printf("i=%i, (%i, %i, %i, %i)\n", i+1, j+1, l+1, l1+1, l2+1);
												//tmp_nu += 2 * build_and_calc_subdetvar(3, old_m, extra_K, extra_q, tmp_K + 1, i, new_n_r, u[l], u[l1], u[l2], u[l3], NULL);
												if (build_and_calc_subdetvar(3, old_m, extra_K, extra_q, tmp_K + 1, i, new_n_r, u[l], u[l1], u[l2], u[l3], NULL) != 0) {
												// ^^ i.e. the intersection of the "unit column vector hyperplane" with the singular locus is non-zero 
												// ==> singular locus is not the empty set
												// ==> euler number preserving effective splitting
												// (tmp_nu is additive, so one non-zero term is enough to conclude it's non-zero. Don't need to continue computing full tmp_nu. Instead break out of calculation at first non-zero term.)
													ineff = false;
													break;
												}
											}
											if (ineff == false) {break;}
										}
										if (ineff == false) {break;}
									}
									if (ineff == false) {break;}
								}
							}
							if (ineff == false) {break;}
							/*printf("tmp_nu=%i\n", tmp_nu);
							if (tmp_nu != 0) { 
							// ^^ i.e. the intersection of the "unit column vector hyperplane" with the singular locus is non-zero 
							// ==> singular locus is not the empty set
							// ==> euler number preserving effective splitting
								ineff = false;
								break;
							}*/
						}
					}
				}
			}
			// free memory
			for (j = 0; j < (old_n_r[i]+1); j++) {
				delete u[j];
			}
			delete u;
			// end: implement (3.24)-(3.25)
			
			new_eulernumber = old_eulernumber - (n-1)*nu;
			if (ineff == true) {
				new_conf = contract(new_m, new_K, new_n_r, new_q, new_eulernumber);
				if (new_conf == NULL) {
					new_conf = (conf*) malloc( sizeof(conf) );
					new_conf->m = new_m;
					new_conf->K = new_K;
					new_conf->eulernumber = new_eulernumber;
					new_conf->n_r = (int*) malloc( sizeof(int)*new_m );
					new_conf->q = (int*) malloc( sizeof(int)*new_m*new_K );
					
					memcpy(new_conf->n_r, new_n_r, sizeof(int)*new_m);
					sort_q(new_m, new_K, new_q);
					memcpy(new_conf->q, new_q, sizeof(int)*new_m*new_K);
				}
				free(new_n_r); free(new_q);
				return new_conf;
			}
			
			free(new_n_r); free(new_q);
		}
	}
	
	return NULL;
}

inline int calc_nu_comp(const int old_m, const int old_K, const int old_q[], const int tmp_K, const int i, const int new_n_r[], const int col1[], const int col2[], const int col3[], const int col4[], const int col5[]) {
	int l1, l2, nu;
	
	int * tmp_q;
	int tmp_row, tmp_col;
	vector<jmonomial> * tmp_mu;
	
	int new_m = old_m - 1;
	
	tmp_q = new int[new_m * tmp_K];
	for (l1 = 0; l1 < old_m; l1++) {
		if (l1 != i) {
			tmp_row = l1 - ((l1 > i) ? 1 : 0);
			tmp_q[0 + (tmp_row * tmp_K)] = col1[tmp_row];
			tmp_q[1 + (tmp_row * tmp_K)] = col2[tmp_row];
			tmp_q[2 + (tmp_row * tmp_K)] = col3[tmp_row];
			tmp_q[3 + (tmp_row * tmp_K)] = col4[tmp_row];
			tmp_q[4 + (tmp_row * tmp_K)] = col5[tmp_row];
			tmp_col = 5;
			for (l2 = 0; l2 < old_K; l2++) {
				if (old_q[l2 + (i*old_K)] == 0) {
					tmp_q[tmp_col + (tmp_row * tmp_K)] = old_q[l2 + (l1*old_K)];
					tmp_col++;
				}
			}
		}
	}
	//debug_output(new_m, tmp_K, new_n_r, tmp_q);
	tmp_mu = new vector<jmonomial>;
	calc_mu(new_m, tmp_K, new_n_r, tmp_q, tmp_mu);
	nu = compute_detvar_eulernumber_fast(0, new_m, tmp_K, new_n_r, tmp_q, tmp_mu);
	clear_mu(tmp_mu);
	delete tmp_mu;
	delete tmp_q;
	
	return nu;
}

inline int build_and_calc_subdetvar(const int my_n, const int old_m, const int old_K, const int old_q[], const int tmp_K, const int i, const int new_n_r[], const int col1[], const int col2[], const int col3[], const int col4[], int * empty) {
	int l1, l2, nu;
	
	int * tmp_q;
	int tmp_row, tmp_col;
	vector<jmonomial> * tmp_mu;
	
	int new_m = old_m - 1;
	
	tmp_q = new int[new_m * tmp_K];
	for (l1 = 0; l1 < old_m; l1++) {
		if (l1 != i) {
			tmp_row = l1 - ((l1 > i) ? 1 : 0);
			tmp_q[0 + (tmp_row * tmp_K)] = col1[tmp_row];
			tmp_q[1 + (tmp_row * tmp_K)] = col2[tmp_row];
			tmp_q[2 + (tmp_row * tmp_K)] = col3[tmp_row];
			tmp_q[3 + (tmp_row * tmp_K)] = col4[tmp_row];
			tmp_col = 4;
			for (l2 = 0; l2 < old_K; l2++) {
				if (old_q[l2 + (i*old_K)] == 0) {
					tmp_q[tmp_col + (tmp_row * tmp_K)] = old_q[l2 + (l1*old_K)];
					tmp_col++;
				}
			}
		}
	}
	//debug_output(new_m, tmp_K, new_n_r, tmp_q);
	tmp_mu = new vector<jmonomial>;
	calc_mu(new_m, tmp_K, new_n_r, tmp_q, tmp_mu);
	nu = compute_detvar_eulernumber_fast(my_n-3, new_m, tmp_K, new_n_r, tmp_q, tmp_mu);
	if ((my_n==4) && (nu == 0)) {
		*empty = is_detvar_emptyset(new_m, tmp_K, new_n_r, tmp_q, tmp_mu);
	}
	clear_mu(tmp_mu);
	delete tmp_mu;
	delete tmp_q;
	
	return nu;
}

// applies all possible identites (in order to find the smallest representation) or NULL if no further identities are applicable
inline conf* apply_identities(const int old_m, const int old_K, const int old_n_r[], const int old_q[], const int old_eulernumber) {
	int i, j, l1, l2, l3;
	int num_P[2*(n+1)];	// array fr. 1 to (2*n+1). num_P[i] is the number of \CP^i factors in n_r. Note: n_r is ordered (ascendingly). num_P[0] is ignored
	int new_m, new_K;
	int * new_n_r, * new_q;
	int new_eulernumber;
	conf* new_conf;
	
	int * tmp_row;
	
	memset(num_P, 0, 2*(n+1)*sizeof(int));
	init_num_P(old_m, old_n_r, num_P);
	
	// start with 1-fold identities II(i) and II(ii)
	
	// II(i): search for [2|2]
	for (i=num_P[1]; i < (num_P[1]+num_P[2]); i++) {
		for (j=0; j < old_K; j++) {
			if ( (old_q[j+(i*old_K)]==2) && (sum_q_column(j,old_m,old_K,old_q)==2) ) {
				new_m = old_m;
				new_K = old_K - 1;
				
				new_n_r = (int*) malloc( sizeof(int)*new_m);
				//new_q = (int*) malloc( sizeof(int)*new_m*new_K);
				
				memcpy(new_n_r, old_n_r, sizeof(int)*new_m);
				new_n_r[num_P[1]]--; // replace a P^2 by a P^1

				class qmat qm (old_m, old_K, old_q); 
				qm.setq(old_m, old_K - 1, qm.delete_col(j));
				qm.setq(old_m, old_K - 1, qm.row_mult(i,2));
				qm.setq(old_m, old_K - 1, qm.move_row_to_top(i));
				new_q = qm.return_qcopy();
				qm.clear();
				
				sort_q(new_m, new_K, new_q);
				//debug_output(old_m, old_K, old_n_r, old_q);
				//debug_output(new_m, new_K, new_n_r, new_q);
				
				new_eulernumber = compute_eulernumber(new_m, new_K, new_n_r, new_q);
				if (new_eulernumber == old_eulernumber) {
					new_conf = apply_identities(new_m, new_K, new_n_r, new_q, new_eulernumber);
					if (new_conf == NULL) {
						new_conf = (conf*) malloc( sizeof(conf) );
						new_conf->m = new_m;
						new_conf->K = new_K;
						new_conf->eulernumber = new_eulernumber;
						new_conf->n_r = (int*) malloc( sizeof(int)*new_m );
						new_conf->q = (int*) malloc( sizeof(int)*new_m*new_K );
						
						memcpy(new_conf->n_r, new_n_r, sizeof(int)*new_m);
						memcpy(new_conf->q, new_q, sizeof(int)*new_m*new_K);
					}
					free(new_n_r); free(new_q);
					return new_conf;
				}
				
				free(new_n_r); free(new_q);
			}
		}
	}
	
	// II(ii): search for [1 1|1 1]^T
	if (num_P[1] >= 2) {
		for (i=0; i < (num_P[1]-1); i++) {
			for (j=i+1; j < num_P[1]; j++) {
				for (l1=0; l1 < old_K; l1++) {
					if ( (old_q[l1+(i*old_K)]==1) && (old_q[l1+(j*old_K)]==1) && (sum_q_column(l1,old_m,old_K,old_q)==2) ) {
						new_m = old_m - 1;
						new_K = old_K - 1;
						
						new_n_r = (int*) malloc( sizeof(int)*new_m);
						//new_q = (int*) malloc( sizeof(int)*new_m*new_K);
						
						memcpy(new_n_r, &old_n_r[1], sizeof(int)*new_m);
						
						// class ops NOT TESTED HERE!
						class qmat qm (old_m, old_K, old_q);
						qm.setq(old_m, old_K - 1, qm.delete_col(l1));
						
						tmp_row = qm.row_addition(i, j);	// == "a+b"
						qm.setq(old_m - 1, old_K - 1, qm.delete_row(j));
						qm.setq(old_m - 2, old_K - 1, qm.delete_row(i));
						qm.setq(old_m - 1, old_K - 1, qm.add_row(tmp_row));
						new_q = qm.return_qcopy();
						delete tmp_row; qm.clear();
						
						sort_q(new_m, new_K, new_q);
						//debug_output(old_m, old_K, old_n_r, old_q);
						//debug_output(new_m, new_K, new_n_r, new_q);
						
						new_eulernumber = compute_eulernumber(new_m, new_K, new_n_r, new_q);
						if (new_eulernumber == old_eulernumber) {
							new_conf = apply_identities(new_m, new_K, new_n_r, new_q, new_eulernumber);
							if (new_conf == NULL) {
								new_conf = (conf*) malloc( sizeof(conf) );
								new_conf->m = new_m;
								new_conf->K = new_K;
								new_conf->eulernumber = new_eulernumber;
								new_conf->n_r = (int*) malloc( sizeof(int)*new_m );
								new_conf->q = (int*) malloc( sizeof(int)*new_m*new_K );
								
								memcpy(new_conf->n_r, new_n_r, sizeof(int)*new_m);
								memcpy(new_conf->q, new_q, sizeof(int)*new_m*new_K);
							}
							free(new_n_r); free(new_q);
							return new_conf;
						}
						
						free(new_n_r); free(new_q);
					}
				}
			}
		}
	}
	
	// III(i): search for [3|2]
	for (i = (num_P[1] + num_P[2]); i < (num_P[1] + num_P[2] + num_P[3]); i++) {
		for (j = 0; j < old_K; j++) {
			if ( (old_q[j+(i*old_K)]==2) && (sum_q_column(j,old_m,old_K,old_q)==2) ) {
				new_m = old_m + 1;
				new_K = old_K - 1;
				
				new_n_r = (int*) malloc( sizeof(int)*new_m);
				new_n_r[0] = 1;
				new_n_r[1] = 1;
				memcpy(&new_n_r[2], old_n_r, sizeof(int)*i);
				if (i != (old_m-1)) {memcpy(&new_n_r[i+2], &old_n_r[i+1], sizeof(int)*(old_m-i-1));}
				
				class qmat qm (old_m, old_K, old_q); 
				qm.setq(old_m, old_K - 1, qm.delete_col(j));
				tmp_row = qm.row(i);
				qm.setq(old_m - 1, old_K - 1, qm.delete_row(i));
				qm.setq(old_m, old_K - 1, qm.add_row(tmp_row));
				qm.setq(old_m + 1, old_K - 1, qm.add_row(tmp_row));
				new_q = qm.return_qcopy();
				delete tmp_row; qm.clear();
				
				sort_q(new_m, new_K, new_q);
				//debug_output(old_m, old_K, old_n_r, old_q);
				//debug_output(new_m, new_K, new_n_r, new_q);
				
				new_eulernumber = compute_eulernumber(new_m, new_K, new_n_r, new_q);
				if (new_eulernumber == old_eulernumber) {
					new_conf = apply_identities(new_m, new_K, new_n_r, new_q, new_eulernumber);
					if (new_conf == NULL) {
						new_conf = (conf*) malloc( sizeof(conf) );
						new_conf->m = new_m;
						new_conf->K = new_K;
						new_conf->eulernumber = new_eulernumber;
						new_conf->n_r = (int*) malloc( sizeof(int)*new_m );
						new_conf->q = (int*) malloc( sizeof(int)*new_m*new_K );
						
						memcpy(new_conf->n_r, new_n_r, sizeof(int)*new_m);
						memcpy(new_conf->q, new_q, sizeof(int)*new_m*new_K);
					}
					free(new_n_r); free(new_q);
					return new_conf;
				}
				
				free(new_n_r); free(new_q);
			}
		}
	}
	
	// III(ii): search for [1 2|2 1]^T
	for (i=0; i < num_P[1]; i++) {
		for (j=num_P[1]; j < (num_P[1]+num_P[2]); j++) {
			for (l1=0; l1 < old_K; l1++) {
				if ( (old_q[l1+(i*old_K)]==2) && (old_q[l1+(j*old_K)]==1) && (sum_q_column(l1,old_m,old_K,old_q)==3) ) {
					new_m = old_m;
					new_K = old_K - 1;
					
					new_n_r = (int*) malloc( sizeof(int)*new_m);
					memcpy(new_n_r, old_n_r, sizeof(int)*new_m);
					new_n_r[num_P[1]]--; // replace a P^2 by a P^1
					
					class qmat qm (old_m, old_K, old_q); 
					qm.setq(old_m, old_K - 1, qm.delete_col(l1));
					tmp_row = qm.row(j); // preserve "a" (the P2 row)
					qm.setq(old_m - 1, old_K - 1, qm.delete_row(j));
					qm.setq(old_m - 2, old_K - 1, qm.delete_row(i));
					qm.setq(old_m - 1, old_K - 1, qm.add_row(tmp_row));
					qm.setq(old_m, old_K - 1, qm.add_row(tmp_row));
					new_q = qm.return_qcopy();
					delete tmp_row; qm.clear();
					
					sort_q(new_m, new_K, new_q);
					//debug_output(old_m, old_K, old_n_r, old_q);
					//debug_output(new_m, new_K, new_n_r, new_q);
					
					new_eulernumber = compute_eulernumber(new_m, new_K, new_n_r, new_q);
					if (new_eulernumber == old_eulernumber) {
						new_conf = apply_identities(new_m, new_K, new_n_r, new_q, new_eulernumber);
						if (new_conf == NULL) {
							new_conf = (conf*) malloc( sizeof(conf) );
							new_conf->m = new_m;
							new_conf->K = new_K;
							new_conf->eulernumber = new_eulernumber;
							new_conf->n_r = (int*) malloc( sizeof(int)*new_m );
							new_conf->q = (int*) malloc( sizeof(int)*new_m*new_K );
							
							memcpy(new_conf->n_r, new_n_r, sizeof(int)*new_m);
							memcpy(new_conf->q, new_q, sizeof(int)*new_m*new_K);
						}
						free(new_n_r); free(new_q);
						return new_conf;
					}
					
					free(new_n_r); free(new_q);
				}
			}
		}
	}
	
	// III(v): search for [2 2|(2 1) (1 1)]^T
	for (i = num_P[1]; i < (num_P[1] + num_P[2] - 1); i++) {
		for (j = i + 1; j < (num_P[1] + num_P[2]); j++) {
			for (l1 = 0; l1 < old_K; l1++) {
				for (l2 = 0; l2 < old_K; l2++) {
					if (l1 != l2) {
						if ( (((old_q[l1+(i*old_K)]==1) && (old_q[l1+(j*old_K)]==1) && 
							   (old_q[l2+(i*old_K)]==1) && (old_q[l2+(j*old_K)]==2)) || 
							  ((old_q[l1+(i*old_K)]==1) && (old_q[l1+(j*old_K)]==1) && 
							   (old_q[l2+(i*old_K)]==2) && (old_q[l2+(j*old_K)]==1))) && 
							   (sum_q_column(l1,old_m,old_K,old_q)==2) && 
							   (sum_q_column(l2,old_m,old_K,old_q)==3) ) {
							new_m = old_m;
							new_K = old_K - 1;
							
							new_n_r = (int*) malloc( sizeof(int)*new_m);
							memcpy(new_n_r, old_n_r, sizeof(int)*new_m);
							new_n_r[num_P[1]]--; // replace a P^2 by a P^1
							
							class qmat qm (old_m, old_K, old_q);
							qm.setq(old_m, old_K, qm.replace_in_col(l2, 1, 2));
							qm.setq(old_m, old_K - 1, qm.delete_col(l1)); // delete (1 1 0 ... 0) column
							if (old_q[l2+(j*old_K)]==2) { // determine which row (i or j) to move to the top
								qm.setq(old_m, old_K - 1, qm.move_row_to_top(j));
							}
							else {
								qm.setq(old_m, old_K - 1, qm.move_row_to_top(i));
							}
							new_q = qm.return_qcopy();
							qm.clear();
							
							sort_q(new_m, new_K, new_q);
							//debug_output(old_m, old_K, old_n_r, old_q);
							//debug_output(new_m, new_K, new_n_r, new_q);
							
							new_eulernumber = compute_eulernumber(new_m, new_K, new_n_r, new_q);
							if (new_eulernumber == old_eulernumber) {
								new_conf = apply_identities(new_m, new_K, new_n_r, new_q, new_eulernumber);
								if (new_conf == NULL) {
									new_conf = (conf*) malloc( sizeof(conf) );
									new_conf->m = new_m;
									new_conf->K = new_K;
									new_conf->eulernumber = new_eulernumber;
									new_conf->n_r = (int*) malloc( sizeof(int)*new_m );
									new_conf->q = (int*) malloc( sizeof(int)*new_m*new_K );
									
									memcpy(new_conf->n_r, new_n_r, sizeof(int)*new_m);
									memcpy(new_conf->q, new_q, sizeof(int)*new_m*new_K);
								}
								free(new_n_r); free(new_q);
								return new_conf;
							}
							
							free(new_n_r); free(new_q);
						}
					}
				}
			}
		}
	}
	
	return NULL;
}