(* ::Package:: *)

(* ::Title:: *)
(*Redundancy test for 4-fold CICY configurations*)


(* ::Section::Closed:: *)
(*Various modules to test equivalence*)


(* ::Input:: *)
(* tests if row and column sets are equal and if trace and eigenvalues are the same *)

NecEquiv[A_, B_, prot_: False] := 
  Module[{dA, dB, As, Bs, ATs, BTs, equiv, eigA, eigB, AAT, BBT, 
    acc},
   acc = 1/1000;
   (* test matrix size *)
   dA = Dimensions[A]; dB = Dimensions[B];
   If[dA != dB, If[prot, Print["Matrices have different size."]]; 
    Return[False]];
   (* first test row and column sets *)
   As = Sort[Map[Sort, A]]; 
   Bs = Sort[Map[Sort, B]];
   ATs = Sort[Map[Sort, Transpose[A]]]; 
   BTs = Sort[Map[Sort, Transpose[B]]];
   equiv = (As == Bs) && (ATs == BTs);
   If[prot, Print["Row and column sets equal: ", equiv]];
   If[! equiv, Return[False]];
   (* if they are the same look at trace *)
   AAT = A.Transpose[A]; 
   BBT = B.Transpose[B];
   equiv = (Tr[AAT] == Tr[BBT]);
   If[prot, Print["Trace equal: ", equiv]];
   If[! equiv, Return[False]];
   (* if trace is also the same look at eigenvalues *)
   
   eigA = Round[Eigenvalues[AAT // N], acc]; 
   eigB = Round[Eigenvalues[BBT // N], acc];
   equiv = (eigA == eigB);
   If[prot, Print["Eigenvalues equal: ", equiv]];
   If[! equiv, Return[False]];
   equiv];


(* ::Input:: *)
(* necessary and sufficient equivalence test *)
Options[SuffEquiv] = {Prot -> False, Acc -> 1/10000, Precis -> 15};
SuffEquiv[A_, B_, opt___] := Module[
   {acc, dA, dB, a, b, aaT, bbT, esa, esb, eiga, eigb, va, vb, na, nb,
     sa, sb, P, Plsts, bp, asort, bpsort, equiv, s, Plst, equivlst, 
    Plst1, i, Ra, Rb, precis, prot, fname, ret},
   (* options *)
   prot = Prot /. {opt} /. Options[SuffEquiv];
   acc = Acc /. {opt} /. Options[SuffEquiv];
   precis = Precis /. {opt} /. Options[SuffEquiv];
   If[prot, Print["Accuracy: ", acc, ", Precision: ", precis]];
   (* test matrix size *)
   dA = Dimensions[A]; dB = Dimensions[B];
   If[dA != dB, If[prot, Print["Matrices have different size."]]; 
    Return[False]];
   (* ensure matrices have more columns than rows *)
   
   If[dA[[1]] > dA[[2]], a = Transpose[A]; b = Transpose[B], a = A; 
    b = B];
   (* find eigensystems *)
   aaT = a.Transpose[a]; 
   bbT = b.Transpose[b];
   esa = Chop[Eigensystem[N[aaT, precis]]]; 
   esb = Chop[Eigensystem[N[bbT, precis]]];
   eiga = esa[[1]]; eigb = esb[[1]]; va = esa[[2]]; vb = esb[[2]];
   equiv = (eiga == eigb);
   If[prot, Print["Eigenvalues equal: ", equiv, ", Values: ", eiga]];
   If[! equiv, Return[False]];
   (* check if spectrum is degenerate *)
   
   If[Length[DeleteDuplicates[eiga]] < Length[eiga],
    Print["Warning: spectrum degenerate", {A, B}];
    fname = 
     ".m" <> ToString[AbsoluteTime[DateString[]]] <> 
      ToString[
       NumberForm[RandomInteger[{1, 9999}], 4, 
        NumberPadding -> {"0", ""}]];
    WriteCStyleList[path <> fname, {A, B}, {0, 0}];
    Print[
     Timing[ret = 
       RunThrough[path <> "equalq 4 " <> path <> fname, ""]]];
    DeleteFile[path <> fname];
    Switch[ret, 0, Return[False], 1, Return[True], _, 
     Print["Error in Run command"]; Return[False]]
    ];
   (* compute norms and signs and compute standard eigenvectors *)
   
   na = Map[Norm, va]; nb = Map[Norm, vb];
   sa = Tuples[
     Map[Sign, Round[Map[Total, va], acc]] /. {1 -> {1}, -1 -> {-1}, 
       0 -> {1}}];
   sb = Tuples[
     Map[Sign, Round[Map[Total, vb], acc]] /. {1 -> {1}, -1 -> {-1}, 
       0 -> {1, -1}}];
   If[prot && (Length[sa] + Length[sb]) > 2, 
    Print["Eigenvector signs not uniquely fixed"]];
   Ra = Table[
     Transpose[
      Table[sa[[s, i]]*va[[i]]/na[[i]], {i, 1, Length[va]}]], {s, 1, 
      Length[sa]}];
   Rb = Table[
     Transpose[
      Table[sb[[s, i]]*vb[[i]]/nb[[i]], {i, 1, Length[vb]}]], {s, 1, 
      Length[sb]}];
   (* compute row rotation and test if it is a permutation *)
   
   Plst = Flatten[
     Table[Map[Round[#, acc] &, Ra[[s]].Transpose[Rb[[t]]], 
       Infinity], {s, 1, Length[Ra]}, {t, 1, Length[Rb]}], 1];
   If[prot, Print["Row rotation(s): ", Map[MatrixForm, Plst]]];
    (* For[s=1,s<=Length[Plst],s++,Print[Plst[[s]]//MM]]; *)
   
   Plsts = Table[Reverse[Sort[Plst[[s]]]], {s, 1, Length[Plst]}];
   equivlst = 
    Table[Plsts[[s]] == 
      IdentityMatrix[Dimensions[Plst[[s]]][[1]]], {s, 1, 
      Length[Plsts]}];
   equiv = Apply[Or, equivlst];
   If[prot, Print["Row rotation is a permutation: ", equiv]];
   If[! equiv, Return[False]];
   (* compute row permutation of B and check if it has the same columns as A *)
   Plst1 = Pick[Plst, equivlst];
   bp = Table[Plst1[[s]].b, {s, 1, Length[Plst1]}];
   asort = Sort[Transpose[a]]; 
   bpsort = Map[Sort, Map[Transpose, bp]];
   equivlst = Table[asort == bpsort[[s]], {s, 1, Length[bpsort]}];
   equiv = Apply[Or, equivlst];
   equiv];


(* ::Input:: *)
(* permute rows and columns of a matrix randomly *)
RCPerm[m_] := Module[{nr, nc, rperm, cperm},
   {nr, nc} = Dimensions[m];
   rperm = System`RandomPermutation[nr];
   cperm = System`RandomPermutation[nc];
   Transpose[
    System`Permute[Transpose[System`Permute[m, rperm]], cperm]]];


(* ::Input:: *)
(* combine the above modules for an efficient necessary and sufficient check *)
Equiv[A_, B_, prot_: False] := Module[{equiv},
   equiv = NecEquiv[A, B, prot];
   If[! equiv, Return[False]];
   SuffEquiv[A, B, Prot -> prot]];


(* ::Section::Closed:: *)
(*Text input in C style format*)


(* ::Input:: *)
ReadCStyleList[filename_]:=Module[{eulernolist={},qlist={},str,inpline,tmpmatrix={}},
str=OpenRead[filename];
Monitor[While[(inpline=Read[str,String])=!=EndOfFile,
Switch[Characters[inpline][[1]],Characters["M"][[1]],AppendTo[qlist,tmpmatrix];tmpmatrix={};,Characters["S"][[1]],True;,Characters["E"][[1]],AppendTo[eulernolist,ToExpression[StringReplace[inpline,"EULERNUMBER :"->""]]];,_,AppendTo[tmpmatrix,Delete[Flatten[ImportString[inpline,"Table"],1],{{1},{2}}]];];
];
,Length[qlist]];
Close[str];
qlist=Delete[qlist,1];
AppendTo[qlist,tmpmatrix];
Return[{qlist,eulernolist}];
];


(* ::Section::Closed:: *)
(*Text output in C style format*)


(* ::Input:: *)
WriteCStyleList[filename_,qlist_,eulernolist_]:=Module[{str,i,j,k,B,m,K},
str=OpenWrite[filename];
For[i=1,i<=Length[qlist],i++,
B=qlist[[i]];m=Dimensions[B][[1]];K=Dimensions[B][[2]];
WriteString[str,"MATRIX NUMBER :", PaddedForm[i,5],"\nSIZE :",PaddedForm[m,3],PaddedForm[K,2],"\nEULERNUMBER :",PaddedForm[eulernolist[[i]],4],"\n"];
For[j=1,j<=m,j++,
WriteString[str,Sum[B[[j,k]],{k,1,K}]-1," |"];
For[k=1,k<=K,k++,
WriteString[str," ",B[[j,k]]];
];
WriteString[str,"\n"];
];
WriteString[str,"\n"];
];
Close[str];
];


(* ::Input:: *)
WriteCStyle1List[filename_, list_] := Module[{str, i, j, k, B, m, K},
   str = OpenWrite[filename];
   For[i = 1, i <= Length[list], i++,
    B = list[[i]][[2]]; m = Dimensions[B][[1]]; K = Dimensions[B][[2]];
    WriteString[str, "MATRIX NUMBER :", PaddedForm[i, 5], "\nSIZE :", 
     PaddedForm[m, 3], PaddedForm[K, 2], "\nEULERNUMBER :", 
     PaddedForm[list[[i]][[1]], 4], "\n"];
    For[j = 1, j <= m, j++,
     WriteString[str, Sum[B[[j, k]], {k, 1, K}] - 1, " |"];
     For[k = 1, k <= K, k++,
      WriteString[str, " ", B[[j, k]]];
      ];
     WriteString[str, "\n"];
     ];
    WriteString[str, "\n"];
    ];
   Close[str];
   ];
   

(* ::Section::Closed:: *)
(*Own modules to test equivalence*)

ModOutEqiv[qlist_, eulernolist_, offset1_: 1, offset2_: 1] := 
  Module[{i, j, redqlist = {}, redeulernolist = {}, eq},
   If[offset2 > 1,
    redqlist = qlist[[1 ;; offset2 - 1]];
    redeulernolist = eulernolist[[1 ;; offset2 - 1]];
    ];
   For[i = offset2, i <= Length[qlist], i++,
    WriteString["stdout", i, ":\t", DateString[], "\n"];
    eq = False;
    For[j = offset1, j < i, j++,
     If[( (Dimensions[qlist[[j]]][[1]] > 
          Dimensions[qlist[[i]]][[1]]) ||   (Dimensions[qlist[[j]]][[
           2]] > Dimensions[qlist[[i]]][[2]]) ), Break[];];
     eq = Equiv[qlist[[j]], qlist[[i]]];
     If[eq, WriteString["stdout", "j: ", j, "  i: ", i, "\n"]; 
      Break[];];
     ];
    If[! eq, AppendTo[redqlist, qlist[[i]]]; 
     AppendTo[redeulernolist, eulernolist[[i]]];];
    ];
   Return[{redqlist, redeulernolist}];
   ];

(* ::Input:: *)
ReOrgList[list_] := 
  Module[{i, j, newlist}, 
   newlist = 
    Flatten[Table[{{i, j}, 
       Select[list, Dimensions[#[[2]]] == {i, j} &]}, {i, 1, 20}, {j, 
       1, 24}], 1];
   newlist = DeleteCases[newlist, {{_, _}, {}}];
   Return[newlist];];

   
(* ::Input:: *)
ModOutEqivSublists[oldlist_, newlist_] := 
  Module[{i, j, k, olreorg, redlist, equiv, newsize}, 
   redlist = oldlist;
   olreorg = ReOrgList[oldlist];
   For[i = 1, i <= Length[newlist], i++, 
    WriteString["stdout", i, ":\t", DateString[], "\n"];
    equiv = False; newsize = True;
    For[j = 1, (j <= Length[olreorg] && newsize), j++, 
     If[Dimensions[newlist[[i]][[2]]] == olreorg[[j]][[1]], 
       newsize = False;
       For[k = 1, (k <= Length[olreorg[[j]][[2]]] && ! equiv), k++, 
        equiv = Equiv[olreorg[[j]][[2]][[k]][[2]], 
          newlist[[i]][[2]]];
        If[equiv, WriteString["stdout", "eq", "\n"];];];
       If[! equiv, AppendTo[redlist, newlist[[i]]];];];];
    If[newsize, AppendTo[redlist, newlist[[i]]];];];
   Return[redlist];];
  
 
(* ::Input:: *)
MergeQandENList[path_, fnameQ_, fnameEN_, fnameNEW_] := 
  Module[{list}, list = {Get[path <> fnameQ], Get[path <> fnameEN]};
   list = 
    Table[{list[[2]][[i]], list[[1]][[i]]}, {i, 1, Length[list[[1]]]}];
   list >> path <> fnameNEW;];
   
   
(* ::Section:: *)
(*Carry out equivalence tests on data sets*)


(* ::Input:: *)
(* set path to data directory *)
path=$HomeDirectory<>"/scr/test/CICY4_full_package121019/";


(* ::Input:: *)
(* find classes of equivalent configurations *)
Print[Timing[lstT1 = Get[path <> "L10prime.m1.txt"]; Length[lstT1]]]
Print[Timing[lstT2 = Get[path <> "L11.m1.txt"]; Length[lstT2]]]
Print[Timing[redlstT12 = ModOutEqivSublists[lstT1, lstT2]; Length[redlstT12]]]

(* ::Input:: *)
WriteCStyle1List[path <> "L11prime_pre.txt", redlstT12];


(* ::Input:: *)
Exit[];