/*
 *  This program reads two CICY lists and checks for differences between the two lists
 *  filename1 == smaller list, filename2 == larger list
 *
 *  Created by Alexander Haupt on 17/05/2010.
 *  Copyright 2013 Alexander Haupt. All rights reserved.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <time.h>
#include <algorithm>

#if defined(__APPLE__) || defined(__MACH__) || defined(__OSX__)
  #include <malloc/malloc.h>
#else
  #include <malloc.h>
#endif

#include "mymemcpy.h"

// whether (1) or not (0) to sort (in ascending order) the columns of each matrix when loaded from txt file
// this is only needed when the ORIG 1988 CICY 3-fold list (or another list where the matrix columns or not ordered properly) is used in filename2
#define SORT_AT_LOAD 1

using namespace std;
using namespace __gnu_cxx;

typedef struct node {
	int m;
	int K;
	int eulernumber;
	int * n_r;
	int * q;
} qnode;

int m_tmp;

void complists(const int n, const char fn1[], const char fn2[], const int num1, const int num2);
inline void debug_output(const int m, const int K, const int n_r[], const int q[]);
inline int is_in_list2(const int matno1, const int num2, const qnode list2[], int m, int K, int n_r[], int q[], int in_list1[]);
inline void sort_q(const int m, const int K, int q[]);
int col_cmp(const void *a, const void *b);


int main (int argc, const char * argv[]) {
	if (argc != 6) {
		// input: n for CICY n-fold; filenames of 2 lists; numi = number of configuration matrix in i-th file
		printf("Usage: %s n filename1 filename2 num1 num2\n", argv[0]);
	}
	else {
		complists(atoi(argv[1]), argv[2], argv[3], atoi(argv[4]), atoi(argv[5]));
	}
	
	return 0;
}

void complists(const int n, const char fn1[], const char fn2[], const int num1, const int num2) {
	int i, linecounter = 0, counter = 0;
	int m, K, eulernumber, orig_matno;
	int * n_r;
	int * q;	// q part of CICY config. matrix
	
	FILE * fp1 = fopen(fn1, "r");
	FILE * fp2 = fopen(fn2, "r");
	
	//qnode list1[num1];
	//qnode list2[num2];
	// ^^^ stack size limited. Instead, allocate in heap:
	qnode *list1, *list2;
	list1 = (qnode*)malloc(num1*sizeof(qnode));
	list2 = (qnode*)malloc(num2*sizeof(qnode));
		
	linecounter = 0; counter = 0;
	if (fp1 != NULL) {
		char line[128], strnum[2] = {0,0};
		
		while(fgets(line, sizeof line, fp1) != NULL) { /* read a line */
			switch (linecounter) {
				case 0:
					if (line[0] == 'M') {
						sscanf(line, "MATRIX NUMBER :%6i\n", &orig_matno);
						linecounter++;
					}
					break;
				case 1:
					sscanf(line, "SIZE :%4i%3i\n", &m, &K);
					n_r = (int*)malloc(m*sizeof(int));
					q = (int*)malloc(m*K*sizeof(int));
					linecounter++;
					//printf("m=%i, K=%i, ", m, K);	// debug
					break;
				case 2:
					sscanf(line, "EULERNUMBER :%5i\n", &eulernumber);
					linecounter++;
					//printf("eulernumber=%i\n", eulernumber);	// debug
					break;
				default:
					if (linecounter <= (m+2)) {
						n_r[linecounter-3] = 0;
						for (i = 0; i < K; i++) {
							strnum[0] = line[(2*i)+4];
							q[i+((linecounter-3)*K)] = atoi(strnum);
							n_r[linecounter-3] += q[i+((linecounter-3)*K)];
						}
						n_r[linecounter-3]--;
						if (linecounter == (m+2)) {
							list1[counter].m = m;
							list1[counter].K = K;
							list1[counter].n_r = n_r;
#if SORT_AT_LOAD == 1
							sort_q(m, K, q);
#endif							
							list1[counter].q = q;
							list1[counter].eulernumber = eulernumber;
							counter++;
						}
						linecounter++;
					}
					else {
						linecounter = 0;
					}
					break;
			}
		}
	}
	else {
		perror(fn1); /* why didn't the file open? */
	}

	linecounter = 0; counter = 0;
	if (fp2 != NULL) {
		char line[128], strnum[2] = {0,0};
		
		while(fgets(line, sizeof line, fp2) != NULL) { /* read a line */
			switch (linecounter) {
				case 0:
					if (line[0] == 'M') {
						sscanf(line, "MATRIX NUMBER :%6i\n", &orig_matno);
						linecounter++;
					}
					break;
				case 1:
					sscanf(line, "SIZE :%4i%3i\n", &m, &K);
					n_r = (int*)malloc(m*sizeof(int));
					q = (int*)malloc(m*K*sizeof(int));
					linecounter++;
					//printf("m=%i, K=%i, ", m, K);	// debug
					break;
				case 2:
					sscanf(line, "EULERNUMBER :%5i\n", &eulernumber);
					linecounter++;
					//printf("eulernumber=%i\n", eulernumber);	// debug
					break;
				default:
					if (linecounter <= (m+2)) {
						n_r[linecounter-3] = 0;
						for (i = 0; i < K; i++) {
							strnum[0] = line[(2*i)+4];
							q[i+((linecounter-3)*K)] = atoi(strnum);
							n_r[linecounter-3] += q[i+((linecounter-3)*K)];
						}
						n_r[linecounter-3]--;
						if (linecounter == (m+2)) {
							list2[counter].m = m;
							list2[counter].K = K;
							list2[counter].n_r = n_r;
#if SORT_AT_LOAD == 1
							sort_q(m, K, q);
#endif														
							list2[counter].q = q;
							list2[counter].eulernumber = eulernumber;
							counter++;
						}
						linecounter++;
					}
					else {
						linecounter = 0;
					}
					break;
			}
		}
	}
	else {
		perror(fn2); /* why didn't the file open? */
	}
	
	fclose(fp1);
	fclose(fp2);
	
	// new 2011
	int in_list1[num2]; // matrix numbers of those configs in list 2 that are also in list 1
	memset(in_list1, 0, sizeof(int)*num2);
	// end new 2011
	
	printf("The following configuration matrices are contained in %s but not in %s:\n\n", fn1, fn2);
	for (i = 0; i < num1; i++) {
		if (!is_in_list2(i, num2, list2, list1[i].m, list1[i].K, list1[i].n_r, list1[i].q, in_list1)) {
			debug_output(list1[i].m, list1[i].K, list1[i].n_r, list1[i].q);
		}
	}
	
	// new 2011	
	printf("\n\nThe following configuration matrices are contained in %s but not in %s:\n\n", fn2, fn1);
	counter = 0;
	for (i = 0; i < num2; i++) {
		if (in_list1[i] == 0) {
			printf("MATRIX NUMBER :%6i (orig: %i)\nSIZE :%4i%3i\nEULERNUMBER :%5i\n", ++counter, i+1, list2[i].m, list2[i].K, list2[i].eulernumber);
			debug_output(list2[i].m, list2[i].K, list2[i].n_r, list2[i].q);
		}
	}
	
	free(list1); free(list2); list1 = NULL; list2 = NULL;
}

inline void debug_output(const int m, const int K, const int n_r[], const int q[]) {
	int i, j;
	
	for (i = 0; i < m; i++) {
		printf("%i |", n_r[i]);
		for (j = 0; j < K; j++) {
			printf(" %i", q[j+(i*K)]);
		}
		printf("\n");
	}
	printf("\n");
}

inline int is_in_list2(const int matno1, const int num2, const qnode list2[], int m, int K, int n_r[], int q[], int in_list1[]) {
	int i, j, validperm;
	int foundmatch = false;
//	qnode * curr_qnode;
	int permuted_q[m*K];
	
	int * owntmpperm = new int[m];
	for (i = 0; i < m; i++) {
		owntmpperm[i] = i;	// set start value: 1st permutation is trivial (0,1,2,...,m-1)
	}
	
	do {
		// we've produced a new permutation, but we need to check that it is a valid one, i.e. rows are only
		// swapped if the dimension of the respective projective spaces is equal
		validperm = true;
		for (i = 0; i < m; i++) {
			if (n_r[owntmpperm[i]] != n_r[i]) {
				validperm = false;
				break;
			}
		}
		if (validperm == false) {continue;}
		
		// generate permuted q-matrix
		for (i = 0; i < m; i++) {
			mymemcpy(&permuted_q[i*K], &q[(owntmpperm[i])*K], K);
		}
		
		// sort columns of permuted q-matrix
		sort_q(m, K, permuted_q);
		
		// walk list2
		for (i = 0; i < num2; i++) {
			if ( (m != list2[i].m) || (K != list2[i].K) ) {
				continue;
			}
			
			for (j = 0; j < (m*K); j++) {
				if (list2[i].q[j] == permuted_q[j]) {
					foundmatch = true;
				}
				else {
					foundmatch = false;
					break;
				}
			}
			
			if (foundmatch == true) {
				break;
			}			
		}
		
		if (foundmatch == true) {
			printf("1: %i --> 2: %i\n", matno1+1, i+1);
			in_list1[i] = 1; // new 2011
			break;
		}
	} while ( next_permutation (owntmpperm,owntmpperm+m) );
	
	delete[] owntmpperm;
	return foundmatch;
}

inline void sort_q(const int m, const int K, int q[]) {
	int i, j;
	int columns[K][m];
	
	// decompose q into K column vectors (K 1-d arrays)
	for (i = 0; i < K; i++) {
		for (j = 0; j < m; j++) {
			columns[i][j] = q[i + (j*K)];
		}
	}
	
	// sort using qsort library function
	m_tmp = m;	// need this to tell col_cmp function how long the column vectors are (i.e. length: m)
	qsort(columns, K, m*sizeof(int), col_cmp);
	
	// rebuild q-matrix out of the ordered column vectors
	for (i = 0; i < K; i++) {
		for (j = 0; j < m; j++) {
			q[i + (j*K)] = columns[i][j];
		}
	}
}

// qsort column array comparison function
int col_cmp(const void *a, const void *b) {
	int i = 0, ret = 0;
	
	const int * ia = (const int *)a; // casting pointer types
	const int * ib = (const int *)b;
	
	while ((i < m_tmp) && ((ret = ia[i] - ib[i]) == 0)) {
		i++;
	}
	
	return ret; // integer comparison: returns negative if b > a and positive if a > b
}