/*
 *  Helper code to compute the Euler characteristic of a CICY configuration.
 *
 *  Created by Alexander Haupt on 08/03/2011.
 *  Copyright 2013 Alexander Haupt. All rights reserved.
 *
 */

#include "euler.h"

// returns eulernumber of CICY
int compute_eulernumber(const int m, const int K, const register int n_r[], const register int q[]) {
//	// as long as it's not implemented for n != 3, return 0
//	if (n != 3) { return 0;}
	
	// for small 3-fold matrices: computer eulernumber directly (in all other cases: call mathematica --> see below)
	if ( (n==3) && ( (m < 9) || (K < 12) ) ) {
		
		int ret = 0, i, j, mu_coeff, pos, qprod;	
		int mu_poly[m];
		memset(mu_poly, 0, sizeof(int)*m);
		mu_poly[m-1] = K;
		
		int idx[K], c3_idx[3];
		int c3_coeff;
		
		do {
			mu_coeff = -1;
			
			// if (J_r)^p with p > n_r[r] appears, the term vanishes
			for (i = 0; i < m; i++) {
				if (mu_poly[i] > n_r[i]) {
					mu_coeff = 0;
					break;
				}
			}
			if (mu_coeff == 0) {
				continue;
			}
			
			// 1. step: calculate coeffiecent of c_3 multiplying into current \mu polynomial
			// this is fast. Only if this coefficient is non-zero, we need to computer the \mu coefficient (which is more complicated)
			
			// we must find the matching part of c_3 to make a top form of the right form degree
			
			// init c3_idx
			pos = 0;
			for (i = 0; i < m; i++) {
				for (j = 0; j < (n_r[i] - mu_poly[i]); j++) {
					c3_idx[pos++] = i;
				}
			}
			
			c3_coeff = 0;
			for (i = 0; i < K; i++) {
				c3_coeff -= q[i + (c3_idx[0])*K] * q[i + (c3_idx[1])*K] * q[i + (c3_idx[2])*K];
			}
			if ((c3_idx[0] == c3_idx[1]) && (c3_idx[1] == c3_idx[2])) {
				c3_coeff += n_r[c3_idx[0]] + 1;
			}
			
			// get combinatorical factors right
			if ((c3_idx[0] != c3_idx[1]) && (c3_idx[1] != c3_idx[2]) && (c3_idx[0] != c3_idx[2])) {
				// all 3 are different. Eg. (1,2,3) --> factor of 3! = 6
				c3_coeff *= 6;
			}
			else if ((c3_idx[0] == c3_idx[1]) || (c3_idx[1] == c3_idx[2]) || (c3_idx[0] == c3_idx[2])) {
				if (!((c3_idx[0] == c3_idx[1]) && (c3_idx[1] == c3_idx[2]))) {
					// 2 are the same. Eg. (1,2,2) --> factor of 3
					c3_coeff *= 3;
				}
			}
			
			// now we have computed the c_3 coefficient, if it is zero we can skip to the next \mu polynomial
			if (c3_coeff == 0) {
				continue;
			}
			
			// 2. step: calculate coeffiecent for \mu polynomial
			
			// init idx-array. Eg. if poly = (1,2,2) --> idx = (1,2,2,3,3)
			pos = 0;
			for (i = 0; i < m; i++) {
				for (j = 0; j < mu_poly[i]; j++) {
					idx[pos++] = i;
				}
			}
			
			mu_coeff = 0;
			do {
				qprod = 1;
				for (i = 0; i < K; i++) {
					qprod *= q[i + (idx[i]*K)];
					if (qprod == 0) {
						break;
					}
				}
				mu_coeff += qprod;
			} while ( next_permutation (idx,idx+K) );
			
			ret += mu_coeff * c3_coeff;
		} while (increase_poly(m, K, mu_poly));
		// loop for each polynomial in J_r in \mu
		
		return ret/n;
	}
	// for large matrices: use mathematica to computer eulernumber
	else {
		FILE *mathfp;
		char line[20], str_cmd_plus_matrix[2600];
		int i, j;
		int result; char *p; bool is_err;
		
#if defined(__APPLE__) || defined(__MACH__) || defined(__OSX__)
		sprintf(str_cmd_plus_matrix, "/Applications/Mathematica.app/Contents/MacOS/MathKernel -noprompt -run \"<<eulerno%i.m\" \"", n);
#else
		sprintf(str_cmd_plus_matrix, "math -noprompt -script eulerno%i.m \"", n);
#endif		
		
		strcat(str_cmd_plus_matrix, "{");
		for (i=0; i < m; i++) {
			strcat(str_cmd_plus_matrix, "{");
			for (j=0; j < K; j++) {
				sprintf(str_cmd_plus_matrix, "%s%i", str_cmd_plus_matrix, q[j+(i*K)]);
				if (j < (K-1)) {
					strcat(str_cmd_plus_matrix, ",");
				}
			}
			strcat(str_cmd_plus_matrix, "}");
			if (i < (m-1)) {
				strcat(str_cmd_plus_matrix, ",");
			}		
		}
		strcat(str_cmd_plus_matrix, "}\"");
		
		// this is a crucial step, so need proper error checking here!
		do {
			is_err = false;
			errno = 0;
			mathfp = popen(str_cmd_plus_matrix, "r");
			if (errno != 0 || mathfp == NULL) {is_err = true; printf("[euler.cpp] ERRNO: %i (popen)\n", errno);}

			if (fgets(line, sizeof line, mathfp) == NULL) {is_err = true; printf("[euler.cpp] ERR: (fgets)\n");}
			if (pclose(mathfp) == -1) {is_err = true; printf("[euler.cpp] ERR: (pclose)\n");}

			errno = 0;
			result = strtol(line, &p, 10);
			if (errno != 0 || *p != 0 || p == line) {is_err = true; printf("[euler.cpp] ERRNO: %i (output: %s)\n", errno, line); debug_output(m, K, n_r, q);}
		} while (is_err == true);
		
		return result;	//printf("%i\n", ret); // debug		
	}
}

// poly: 003 --> 012 --> 021 --> 030 --> ... --> 300, return false if it reached the last term (here: 300)
int increase_poly(const int m, const int K, register int poly[]) {
	if (poly[0] == K) {
		return false;
	}
	
	int i = m-1;
	while (poly[i] == 0) {
		i--;
	}
	poly[i-1]++;
	poly[m-1] = poly[i] - 1;
	memset(&poly[i], 0, sizeof(int)*(m-1-i));
	
	return true;
}

inline void debug_output(const int m, const int K, const int n_r[], const int q[]) {
	int i, j;
	
	for (i = 0; i < m; i++) {
		printf("%i |", n_r[i]);
		for (j = 0; j < K; j++) {
			printf(" %i", q[j+(i*K)]);
		}
		printf("\n");
	}
	printf("\n");
}