/*
 *  Parallelization of the splitting procedure. It takes a list of k matrices and creates k lists of 1 matrix each. Then "splitting" is executed k times in parallel.
 *
 *  Created by Alexander Haupt on 11/06/2012.
 *  Copyright 2013 Alexander Haupt. All rights reserved.
 *
 */

#if defined(_WIN32) || defined(_WIN64)
  #error not tested on windows
#endif

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <time.h>
#include <set>
#include <vector>
#include <algorithm>
#include <errno.h>

#if defined(__APPLE__) || defined(__MACH__) || defined(__OSX__)
  #include <malloc/malloc.h>
#else
  #include <malloc.h>
#endif

using namespace std;
using namespace __gnu_cxx;

struct conf {
	int m;
	int K;
	int * n_r;
	int * q;
	int eulernumber;
};

int n, m_tmp;

void proc_cicy_list(const char filename[]);
static inline void write_output1(FILE * fp, const int counter, const int m, const int K, const register int n_r[], const register int q[], const int eulernumber);
inline void debug_output(const int m, const int K, const int n_r[], const int q[]);
inline void sort_q(const int m, const int K, int q[]);
int col_cmp(const void *a, const void *b);
inline void add_conf(set<conf> * conflist, const int m, const int K, const int n_r[], const int q[], const int eulernumber);
bool operator <(conf a, conf b);


int main (int argc, const char * argv[]) {
	if (argc != 3) {
		// input: n for CICY n-fold
		printf("Usage: %s n filename\n", argv[0]);
	}
	else {
		n = atoi(argv[1]);
		proc_cicy_list(argv[2]);
	}
	
	return 0;
}

void proc_cicy_list(const char filename[]) {
	int i, linecounter = 0, counter = 0;
	int m, K, eulernumber, orig_matno;
	int * n_r;
	int * q;		// q part of CICY config. matrix
	int num_P[2*(n+1)];	// array fr. 1 to (2*n+1). num_P[i] is the number of \CP^i factors in n_r. Note: n_r is ordered (ascendingly). num_P[0] is ignored
	
	FILE * fp = fopen(filename, "r");

	set<conf> conflist;
	set<conf>::iterator it;
	
	if (fp != NULL) {
		char line[128], strnum[2] = {0,0};
		
		while(fgets(line, sizeof line, fp) != NULL) { /* read a line */
			switch (linecounter) {
				case 0:
					if (line[0] == 'M') {
						sscanf(line, "MATRIX NUMBER :%6i\n", &orig_matno);
						linecounter++;
					}
					break;
				case 1:
					sscanf(line, "SIZE :%4i%3i\n", &m, &K);
					n_r = (int*)malloc(m*sizeof(int));
					q = (int*)malloc(m*K*sizeof(int));
					linecounter++;
					//printf("m=%i, K=%i, ", m, K);	// debug
					break;
				case 2:
					sscanf(line, "EULERNUMBER :%5i\n", &eulernumber);
					linecounter++;
					//printf("eulernumber=%i\n", eulernumber);	// debug
					break;
				default:
					if (linecounter <= (m+2)) {
						n_r[linecounter-3] = 0;
						for (i = 0; i < K; i++) {
							strnum[0] = line[(2*i)+4];
							q[i+((linecounter-3)*K)] = atoi(strnum);
							n_r[linecounter-3] += q[i+((linecounter-3)*K)];
						}
						n_r[linecounter-3]--;
						// only keep CICYs which are embedded in ambient spaces that don't contain P1s
						/*if ( (linecounter == 3) && (n_r[linecounter-3] == 1) ) {
							linecounter = 0;
							continue;
						}*/
						if (linecounter == (m+2)) {
							// ASSUME the is_already_there check has been done elsewhere:
							//if (is_already_there(&conflist, m, K, n_r, q) == conflist.end()) {
								add_conf(&conflist, m, K, n_r, q, eulernumber);
							//}
						}
						linecounter++;
					}
					else {
						linecounter = 0;
					}
					break;
			}
		}
		fclose(fp);
	}
	else {
		perror(filename); /* why didn't the file open? */
	}

	counter = 0;
	for ( it=conflist.begin(); it != conflist.end(); it++ ) {
		char output_filename[50];
		sprintf(output_filename, "fs_%i.txt", ++counter);
		FILE * fp1 = fopen(output_filename, "a");	
		write_output1(fp1, counter, (*it).m, (*it).K, (*it).n_r, (*it).q, (*it).eulernumber);
		fclose(fp1);
		
		char str_cmd[2600];
		sprintf(str_cmd, "./splitting %i 1 %s", n, output_filename);
		
		system(str_cmd);
	}
}

static inline void write_output1(FILE * fp, const int counter, const int m, const int K, const register int n_r[], const register int q[], const int eulernumber) {
	int i, j;
	
	fprintf(fp, "MATRIX NUMBER :%6i\nSIZE :%4i%3i\nEULERNUMBER :%5i\n", counter, m, K, eulernumber);
	for (i=0; i < m; i++) {
		fprintf(fp, "%i |", n_r[i]);
		for (j=0; j < K; j++) {
			fprintf(fp, " %i", q[j+(i*K)]);
		}
		fprintf(fp, "\n");
	}
	fprintf(fp, "\n");
}

inline void debug_output(const int m, const int K, const int n_r[], const int q[]) {
	int i, j;
	
	for (i = 0; i < m; i++) {
		printf("%i |", n_r[i]);
		for (j = 0; j < K; j++) {
			printf(" %i", q[j+(i*K)]);
		}
		printf("\n");
	}
	printf("\n");
}

inline void sort_q(const int m, const int K, int q[]) {
	int i, j;
	int columns[K][m];
	
	// decompose q into K column vectors (K 1-d arrays)
	for (i = 0; i < K; i++) {
		for (j = 0; j < m; j++) {
			columns[i][j] = q[i + (j*K)];
		}
	}
	
	// sort using qsort library function
	m_tmp = m;	// need this to tell col_cmp function how long the column vectors are (i.e. length: m)
	qsort(columns, K, m*sizeof(int), col_cmp);
	
	// rebuild q-matrix out of the ordered column vectors
	for (i = 0; i < K; i++) {
		for (j = 0; j < m; j++) {
			q[i + (j*K)] = columns[i][j];
		}
	}
}

// qsort column array comparison function
int col_cmp(const void *a, const void *b) {
	int i = 0, ret = 0;
	
	const int * ia = (const int *)a; // casting pointer types
	const int * ib = (const int *)b;
	
	while ((i < m_tmp) && ((ret = ia[i] - ib[i]) == 0)) {
		i++;
	}
	
	return ret; // integer comparison: returns negative if b > a and positive if a > b
}

inline void add_conf(set<conf> * conflist, const int m, const int K, const int n_r[], const int q[], const int eulernumber) {
	conf new_conf;
	
	new_conf.m = m;
	new_conf.K = K;
	new_conf.eulernumber = eulernumber;

	new_conf.n_r = (int*) malloc( sizeof(int)*m );	
	memcpy(new_conf.n_r, n_r, sizeof(int)*m);	
	
	new_conf.q = (int*) malloc( sizeof(int)*(m*K) );	
	memcpy(new_conf.q, q, sizeof(int)*(m*K));
	
	conflist->insert(new_conf);
}

bool operator <(conf a, conf b) {
	int i;
	
	if ( a.m<b.m ) return true;
	if ( a.m>b.m ) return false;

	for(i=0; i<(a.m); i++) {
		if ( a.n_r[i]<b.n_r[i] ) return true;
		if ( a.n_r[i]>b.n_r[i] ) return false;
	}	
	
	if ( a.K<b.K ) return true;
	if ( a.K>b.K ) return false;
	
	for(i=0; i<((a.m)*(a.K)); i++) {
		if ( a.q[i]<b.q[i] ) return true;
		if ( a.q[i]>b.q[i] ) return false;
	}
	return false;
}