/*
 *  This program generates a list of all possible ambient spaces.
 *
 *  Created by Alexander Haupt on 14/01/2010.
 *  Copyright 2013 Alexander Haupt. All rights reserved.
 *
 */

#include "noP1list.h"
#include "gen_ambsp_list.h"


// For a given choice of n, generates all possible ambient spaces
void gen_ambsp_list() {
	int m = 1;			// number of complex projective spaces
	int K = 1;			// number of polynomial eqs.
	int * n_r;			// dim of each proj. space
	asnode * last_asnode = NULL;

	for (m; m <= maxd; m++) {
		n_r = init_array(m, 1);
		do {
			// Uncomment this if-statement to restrict to CICYs with n_r > 1 for all r (so called "noP1 CICYs")
			if (n_r[0] == 1) {
				continue;
			}
			K = dim(m, n_r)-n;
			if (K < 1) {
				continue;
			}
			if (is_s_alpha_ineq(m, n_r)) {
				continue;
			}
			
			last_asnode = add_asnode(last_asnode, m, K, n_r);
		} while (increase_dim(m, n_r));
		free(n_r);
		n_r = NULL;
	}
}

// returns false if dimension of ambient space cannot be increased further (i.e. it'd be larger than maxd)
static inline int increase_dim(const int m, register int n_r[]) {
    int i;
	
	for (i=m;  i > 0; i--) {
		n_r[i-1]++;
		
		if (dim(m, n_r) <= maxd) {
			return true;
		}
		else {
			if (i == 1) {
				return false;
			}
			init_val(m-i+2, &n_r[i-2], n_r[i-2]+1);
			if (dim(m, n_r) <= maxd) {
				return true;
			}
		}
		
	}
	return false;
}

// implement constraints p <= \alpha and s <= 3n - (\alpha - p)
// returns true if s > 3n - (\alpha - p)
static inline int is_s_alpha_ineq(const int m, const register int n_r[]) {
	int j, p;
	int s = 0, alpha = 0;
	
	while ( (s < m) && (n_r[s] == 1) ) {
		s++;
	}
	p = m - s; // s: number of P1s, p: number of Pns (with n>1)
	if (p > 0) {
		for (j=s; j<m; j++) {
			alpha += n_r[j];
		}
		alpha -= p;
	}
	if ( (s > ((3*n)-(alpha-p))) || (alpha > (2*n)) ) { // NOTE: p<=\alpha trivially/always satisfied
		return true;
	}
	
	return false;
}

asnode * add_asnode(asnode * last_asnode, const int m, const int K, const register int n_r[]) {
	asnode * new_asnode;
	new_asnode = (asnode*) malloc( sizeof(asnode) );
	new_asnode->m = m;
	new_asnode->K = K;
	new_asnode->n_r = (int*) malloc( sizeof(int)*m );
	
	mymemcpy(new_asnode->n_r, n_r, m);
	new_asnode->next = NULL;
	
	if (last_asnode != NULL) {
		last_asnode->next = new_asnode;
	}
	else {
		as1 = new_asnode;
	}
	
	return new_asnode;
}

void free_aslist() {
	asnode * curr_asnode = as1;
	asnode * tmp = NULL;
	
	while (curr_asnode != NULL) {
		tmp = curr_asnode->next;
		free(curr_asnode->n_r);
		free(curr_asnode);
		curr_asnode = tmp;
	}
	as1 = NULL;
}