/*
 *  This is a special algorithm to create the no P1 list where the ambient space consists of only an even number of P2s.
 *  The idea is to split the q-matrix vertically in halves, i.e. to consider the junction of two m*(K/2) matrices instead of one m*K matrix.
 *  This provides a vast decrease in running time as compared to the "naive algorithm".
 *
 *  Created by Alexander Haupt on 17/06/2011.
 *  Copyright 2013 Alexander Haupt. All rights reserved.
 *
 */

#include "half_config.h"

int n;		// complex dim of CICY (= CICY n-fold)
int maxd;	// maximum dimension of ambient space
int permlistcounter; int * tmpperm; int m_tmp; int glob_q_size1, glob_q_size2, glob_q_size_fullK;


int main (int argc, const char * argv[]) {
	int m, K;
	int * n_r;

	if (argc != 3) {
		printf("Usage: %s n num_p2\n", argv[0]);
	}
	else {
		n = atoi(argv[1]);
		maxd = 7*n;

		m = atoi(argv[2]) + 1;
		K = (2*(m-1)) + 3 - n;
		n_r = init_array(m, 2);
		n_r[m-1]++;

		// note: if (((K/2)*2) == K) then K is even, otherwise it is odd
		// we will assume K is odd

		// note also: this code was written and tested specifically for the input parameters "4 6", i.e. P2^6 x P3
		// or more generally for ambient spaces of the form P2^(m-1) x P3

		halfconfig_enumcicy(m, K, n_r);
	}

	return 0;
}

// Implements algorithm described in section 2.3.1 of T. Huebsch, "Calabi-Yau Manifolds -- A Bestiary for Physicists"
void halfconfig_enumcicy(const int m, const int K, register int n_r[]) {
 	int i, j;

	int K1;
	int subconfig1[5*((m + 2)*(m + 1)*m)/6][m];
	set<qnode> matrixlist1[5*((m + 2)*(m + 1)*m)/6];
	int * tmp_subconfig1, tmp_subconfig_sum1;
	int * q1_start, * q1_end, * subconfig_n_r1;

	int K2;
	int subconfig2[5*((m + 2)*(m + 1)*m)/6][m];
	set<qnode> matrixlist2[5*((m + 2)*(m + 1)*m)/6];
	int * tmp_subconfig2, tmp_subconfig_sum2;
	int * q2_start, * q2_end, * subconfig_n_r2;

	int * sister_subconfig;

	set<qnode> outputlist;

	register int * q;		// q part of CICY config. matrix
	register int counter = 0;	// count number of CICY configs.
	int * num_P;			// array fr. 0 to (n_r[m-1]-1). num_P[i] is the number of CP^{i+1} factors in n_r. Note: n_r is ordered (ascendingly)
	int **n_r_perms;		// array holding all allowed permutations of n_r. E.g. for n_r={2,2,3} it'd be {{1,2,3},{2,1,3}}
	int num_perms;
	// NOTE: don't use global variables for the above (would be slower)

	char filename[50]; char time_str[20]; time_t timer;
	timer = time(NULL);
	strftime(time_str, 50, "%y%m%d_%H%M%S", localtime(&timer));
	sprintf(filename, "output_%s.txt", time_str);
	FILE * fp = fopen(filename, "a");
	
	struct timeval startTime;
	struct timeval endTime;
	gettimeofday(&startTime, NULL);
	struct timeval ru_startTime;
	struct timeval ru_endTime;
	struct rusage ru;
	getrusage(RUSAGE_SELF, &ru);
	ru_startTime = ru.ru_utime;

	K1 = K/2;
	K2 = K - K1;

	glob_q_size1 = m*K1;
	glob_q_size2 = m*K2;
	glob_q_size_fullK = m*K;
	q = init_array(m*K, 0);

	num_P = init_array(n_r[m-1], 0);
	init_num_P(m, n_r, num_P);
	num_perms = calc_num_perms(num_P, n_r[m-1]);
	n_r_perms = init_permlist(m, n_r, num_P, num_perms);

	// init subconfigs 1
	tmp_subconfig1 = init_array(m, 0);
	subconfig_n_r1 = init_array(m, 0);
	q1_start = init_array(m*K1, 0);
	q1_end = init_array(m*K1, 0);
	for (i = 0; i < (5*((m + 2)*(m + 1)*m)/6); i++) {
		mymemcpy(subconfig1[i], tmp_subconfig1, m);

		// find next subconfig
		if (tmp_subconfig1[m-1] < 4) {
			tmp_subconfig1[m-1]++;
		}
		else {
			j = m-2;
			while (tmp_subconfig1[j] == 3) {
				j--;
			}
			tmp_subconfig1[j]++;
			if (j != (m-2)) {
				init_val((m-2)-j, &tmp_subconfig1[j+1], tmp_subconfig1[j]);
			}
			tmp_subconfig1[m-1] = 0;
		}
		// END: find next subconfig

		// debug
		//printf("%i : ", i);
		//for (j = 0; j < m; j++) {
		//	printf("%i", subconfig1[i][j]);
		//}
		//printf("\n");
		// END: debug

		tmp_subconfig_sum1 = 0;
		for (j = 0; j < m; j++) {
			tmp_subconfig_sum1 += subconfig1[i][j];
		}
		if (tmp_subconfig_sum1 < (K1*2)) {	// entries too small to lead to non-trivial subconfig (i.e. sum of each column >= 2 would be violated)
			continue;
		}
		if (tmp_subconfig_sum1 > ((3*m) + 1 - (2*K2))) {	// entries of sister subconfig too small to lead to non-trivial subconfig (i.e. sum of each column >= 2 would be violated)
			continue;
		}

		for (j = 0; j < m; j++) {
			subconfig_n_r1[j] = subconfig1[i][j] - 1;
			q1_start[(j*K1) - 1 + K1] = subconfig1[i][j];	// fill last column
			q1_end[j*K1] = subconfig1[i][j];		// fill first column
		}
		//debug_output(m, K1, subconfig_n_r1, q1_start);
		//debug_output(m, K1, subconfig_n_r1, q1_end);

		enumcicy(m, K1, subconfig_n_r1, q1_start, q1_end, &matrixlist1[i]);

		// debug
		/*int counter1 = 0;
		char fn[50];
		sprintf(fn, "output_%i.txt", i);
		FILE * fp1 = fopen(fn, "a");
		set<qnode>::iterator it;
		for (it = matrixlist1[i].begin(); it != matrixlist1[i].end(); it++) {
			write_output1(fp1, ++counter1, m, K1, subconfig_n_r1, (*it).q, 0);
		}
		fclose(fp1);*/
		// END: debug
	}
	
	// init subconfigs 2
	tmp_subconfig2 = init_array(m, 0);
	subconfig_n_r2 = init_array(m, 0);
	q2_start = init_array(m*K2, 0);
	q2_end = init_array(m*K2, 0);
	for (i = 0; i < (5*((m + 2)*(m + 1)*m)/6); i++) {
		mymemcpy(subconfig2[i], tmp_subconfig2, m);

		// find next subconfig
		if (tmp_subconfig2[m-1] < 4) {
			tmp_subconfig2[m-1]++;
		}
		else {
			j = m-2;
			while (tmp_subconfig2[j] == 3) {
				j--;
			}
			tmp_subconfig2[j]++;
			if (j != (m-2)) {
				init_val((m-2)-j, &tmp_subconfig2[j+1], tmp_subconfig2[j]);
			}
			tmp_subconfig2[m-1] = 0;
		}
		// END: find next subconfig

		// debug
		//printf("%i : ", i);
		//for (j = 0; j < m; j++) {
		//	printf("%i", subconfig2[i][j]);
		//}
		//printf("\n");
		// END: debug

		tmp_subconfig_sum2 = 0;
		for (j = 0; j < m; j++) {
			tmp_subconfig_sum2 += subconfig2[i][j];
		}
		if (tmp_subconfig_sum2 < (K2*2)) {	// entries too small to lead to non-trivial subconfig (i.e. sum of each column >= 2 would be violated)
			continue;
		}
		if (tmp_subconfig_sum2 > ((3*m) + 1 - (2*K1))) {	// entries of sister subconfig too small to lead to non-trivial subconfig (i.e. sum of each column >= 2 would be violated)
			continue;
		}

		for (j = 0; j < m; j++) {
			subconfig_n_r2[j] = subconfig2[i][j] - 1;
			q2_start[(j*K2) - 1 + K2] = subconfig2[i][j];	// fill last column
			q2_end[j*K2] = subconfig2[i][j];		// fill first column
		}
		//debug_output(m, K2, subconfig_n_r2, q2_start);
		//debug_output(m, K2, subconfig_n_r2, q2_end);

		enumcicy(m, K2, subconfig_n_r2, q2_start, q2_end, &matrixlist2[i]);

		// debug
		/*int counter1 = 0;
		char fn[50];
		sprintf(fn, "output_%i.txt", i);
		FILE * fp1 = fopen(fn, "a");
		set<qnode>::iterator it;
		for (it = matrixlist2[i].begin(); it != matrixlist2[i].end(); it++) {
			write_output1(fp1, ++counter1, m, K2, subconfig_n_r2, (*it).q, 0);
		}
		fclose(fp1);*/
		// END: debug
	}

	sister_subconfig = init_array(m, 0);
	for (i = 0; i < (5*((m + 2)*(m + 1)*m)/6); i++) {
		if (matrixlist1[i].empty()) {
			continue;
		}
		
		// find "sister" subconfig
		for (j = 0; j < (m-1); j++) {
			sister_subconfig[j] = 3 - subconfig1[i][(m-2)-j];
		}
		sister_subconfig[m-1] = 4 - subconfig1[i][m-1];

		// debug
		//for (j = 0; j < m; j++) { printf("%i", subconfig1[i][j]); }
		//printf("\n");
		//for (j = 0; j < m; j++) { printf("%i", sister_subconfig[j]); }
		//printf("\n\n");
		// END: debug

		int j1, j2, j3;
		for (j1 = 0; j1 < (5*((m + 2)*(m + 1)*m)/6); j1++) {
			for (j = 0; j < m; j++) {
				if (sister_subconfig[j] != subconfig2[j1][j]) {
					break;
				}
			}
			if (j == m) {
				// j1 is the sister config index
				//printf("%i -- %i\n", i, j1);
				set<qnode>::iterator it1, it2;
				for (it1 = matrixlist1[i].begin(); it1 != matrixlist1[i].end(); it1++) {
					for (it2 = matrixlist2[j1].begin(); it2 != matrixlist2[j1].end(); it2++) {
						// copy both halves of (*it1).q and (*it2).q into q
						for (j2 = 0; j2 < m; j2++) {
							for (j3 = 0; j3 < K1; j3++) {
								q[(j2*K) + j3] = (*it1).q[(j2*K1) + j3];
							}
							for (j3 = 0; j3 < K2; j3++) {
								if (j2 != (m-1)) {
									q[(j2*K) + j3 + K1] = (*it2).q[(((m-2)-j2)*K2) + j3];
								}
								else {
									q[(j2*K) + j3 + K1] = (*it2).q[(j2*K2) + j3];
								}
							}
						}
						sort_q(m, K, q);

						if (!is_already_there(&outputlist, m, K, n_r, q, num_perms, n_r_perms)) {
							add_qnode(&outputlist, q, m*K);
							write_output1(fp, ++counter, m, K, n_r, q, (n<4) ? compute_eulernumber(m, K, n_r, q) : 0);
						}
					}
				}
				break;
			}
		}
		// END: find "sister" subconfig
	}

	gettimeofday(&endTime, NULL);
	double tS = startTime.tv_sec*1000000 + (startTime.tv_usec);
	double tE = endTime.tv_sec*1000000  + (endTime.tv_usec);
	getrusage(RUSAGE_SELF, &ru);
	ru_endTime = ru.ru_utime;
	double ru_tS = ru_startTime.tv_sec*1000000 + (ru_startTime.tv_usec);
	double ru_tE = ru_endTime.tv_sec*1000000  + (ru_endTime.tv_usec);
	printf("Time elapsed: %f seconds, getrusage: %f seconds\n", (tE - tS)/1000000, (ru_tE - ru_tS)/1000000);
	// NOTE: can also use in command line: > time programm prog_arguments
	
	fprintf(fp, "\n----------------------------\ncounter = %i, n = %i, max = %i, runtime = %f s, getrusage = %f s", counter, n , maxd, (tE - tS)/1000000, (ru_tE - ru_tS)/1000000);
	fclose(fp);
}

// Implements algorithm described in section 2.3.1 of T. Huebsch, "Calabi-Yau Manifolds -- A Bestiary for Physicists"
void enumcicy(const int m, const int K, register int n_r[], const int q1[], const int q2[], set<qnode> * matrixlist) {
	register int * q;	// q part of CICY config. matrix
		
	q = init_array(m*K, 0);
	mymemcpy(q, q1, m*K);
	
	do {
		if (!is_nontrivial(m, K, q)) {
			continue;
		}
		if (!is_column_ordered(m, K, q)) {
			continue;
		}
		add_qnode(matrixlist, q, m*K);
	} while ( (memcmp(q, q2, sizeof(int)*m*K) != 0) && (increase_q(m, K, n_r, q)) );

	free(q); q = NULL;
}

// returns true if all polynomial eqs. of configuration matrix are non-trivial
static inline int is_nontrivial(const int m, const int K, const int q[]) {
	int i;
	
	for (i=0; i < K; i++) {
		if (sum_q_column(i, m, K, q) < 2) {
			return false;
		}
	}
	
	return true;
}

static inline void write_output(FILE * fp, const int counter, const int m, const int n_r[]) {
	int i;
	char * output_string = NULL;
	
	fprintf(fp, "%04i : %02i : ", counter, 0);
	
	for (i=0; i < m; i++) {
		if (output_string == NULL) {
			asprintf(&output_string, "(%i", n_r[i]);
		}
		else {
			asprintf(&output_string, "%s,%i", output_string, n_r[i]);
		}
	}
	asprintf(&output_string, "%s)", output_string);
	
	fprintf(fp, "%s\n", output_string);
	
	free(output_string);
	output_string = NULL;
}

static inline void write_output1(FILE * fp, const int counter, const int m, const int K, const register int n_r[], const register int q[], const int eulernumber) {
	int i, j;

	fprintf(fp, "MATRIX NUMBER :%6i\nSIZE :%4i%3i\nEULERNUMBER :%5i\n", counter, m, K, eulernumber);
	for (i=0; i < m; i++) {
		fprintf(fp, "%i |", n_r[i]);
		for (j=0; j < K; j++) {
			fprintf(fp, " %i", q[j+(i*K)]);
		}
		fprintf(fp, "\n");
	}
	fprintf(fp, "\n");
}

// TO IMPORT INTO iWorks '09 Numbers: RENAME OUTPUT FILE AS .CSV FILE, DRAG&DROP CSV FILE ONTO iWorks '09 Numbers Spreadsheet
static inline void write_output_CSV(FILE * fp, const int counter, const int m, const int K, const register int n_r[], const register int q[]) {
	int i, j;
	
	fprintf(fp, "%i,\"", counter);
	
	for (i=0; i < m; i++) {
		fprintf(fp, "%i |", n_r[i]);
		for (j=0; j < K; j++) {
			fprintf(fp, " %i", q[j+(i*K)]);
		}
		if (i != (m-1)) {
			fprintf(fp, "\x0D");
		}
	}
	fprintf(fp, "\",\x0D");
}

void write_output2(const int len, const int q[]) {
	int i;
	char * output_string = NULL;
	
	for (i=0; i < len; i++) {
		if (output_string == NULL) {
			asprintf(&output_string, "\t(%i", q[i]);
		}
		else {
			asprintf(&output_string, "%s,%i", output_string, q[i]);
		}
	}
	asprintf(&output_string, "%s)", output_string);
	
	FILE * fp = fopen("output.txt", "a");
	fprintf(fp, "%s\n", output_string);
	fclose(fp);
	
	free(output_string);
	output_string = NULL;
}

static inline void write_to_stdout(const int m, const int K, const int n_r[], const int q[]) {
	int i, qlen = m*K;
	
	putchar(m);
	putchar(K);
	for (i = 0; i < m; i++) {
		putchar(n_r[i]);
	}
	
	for (i = 0; i < qlen; i++) {
		putchar(q[i]);
	}
	//fflush(stdout);	// <-- try if this speeds things up
}

inline void debug_output(const int m, const int K, const int n_r[], const int q[]) {
	int i, j;
	
	for (i = 0; i < m; i++) {
		printf("%i |", n_r[i]);
		for (j = 0; j < K; j++) {
			printf(" %i", q[j+(i*K)]);
		}
		printf("\n");
	}
	printf("\n");
}

inline void add_qnode(set<qnode> * matrixlist, const int new_q[], const int len) {
	qnode new_qnode;
	new_qnode.len = len;
	new_qnode.q = (int*) malloc( sizeof(int)*len );	
	memcpy(new_qnode.q, new_q, sizeof(int)*len);
	
	matrixlist->insert(new_qnode);
}

bool operator <(qnode a, qnode b)
{
	for(int i=0; i<a.len; i++) {
			if ( a.q[i]<b.q[i] ) return true;
			if ( a.q[i]>b.q[i] ) return false;
	}
	return false;
}

// returns the sum of each entry in the (a+1)-th column of q-matrix (a = 1...K)
static inline int sum_q_column(const int a, const int m, const int K, const int q[]) {
	int i = 0, tmp = 0;
	
	for (i=0; i<m; i++) {
		tmp += q[(i*K) + a];
	}
	
	return tmp;
}

static inline void set_smallest_q(const int m, const int K, const int n_r[], int q[]) {
	int i=0;

	for (i=0; i < m; i++) {
		q[(i*K)+(K-1)] = n_r[i]+1;
	}
}

// returns false if q-matrix cannot be increased further
int increase_q(const int m, const int K, const int n_r[], register int q[]) {
	int i = 0, j, l, greater; register int row;
	
	for (i=0; i < m; i++) {
		row = i*K;
		if (q[row] == (n_r[i] + 1)) {
			// can't increase this row further
			continue;	// "break;" doesn't work!!
		}
		
		j = K-1;
		while (q[j+row] == 0) {
			j--;
		}
		if ( (i < (m-1)) && (n_r[i] == n_r[i+1]) ) {
			greater = false;
			
			for (l=0; l < K; l++) {
				if (l < (j-1)) {
					if (q[l+row] > q[l+row+K]) {
						break;
					}
					else if (q[l+row] < q[l+row+K]) {
						greater = true;
					}
					continue;
				}
				else if (l == (j-1)) {
					if ((q[l+row]+1) > q[l+row+K]) {
						break;
					}
					else if ((q[l+row]+1) < q[l+row+K]) {
						greater = true;
					}
					continue;
				}
				else {	// (l > (j-1))
					if (l == (K-1)) {
						if ((q[j+row]-1) > q[l+row+K]) {
							break;
						}
						else if ((q[j+row]-1) < q[l+row+K]) {
							greater = true;
						}
						continue;
					}
					else { // (l < (K-1))
						if (0 < q[l+row+K]) {
							greater = true;
						}
						continue;
					}					
				}
			}
			
			if ( (l<K) && (greater==false) ) {
				continue;
			}
		}
		// no changes to current q-matrix up to here
		q[j-1+row]++;
		q[K-1+row] = q[j+row] - 1;
		init_val(K-j-1, &q[j+row], 0);
		if (i > 0) {
			init_val(row, q, 0);
			set_smallest_q(i, K, n_r, q);
		}
		return true;
	}
	
	return false;
}

// returns true if a-th column is <= than (a+1) column (treated as integer)
static inline int column_cmp(const int a, const int m, const int K, const register int q[]) {
	int i = 0, greater = false;
	
	while ((i < m) && (q[(a-1)+(i*K)] <= q[a+(i*K)])) {
		if (q[(a-1)+(i*K)] < q[a+(i*K)]) {
			greater = true;
		}
		i++;
	}
	
	if ( (i<m) && (greater==false) ) {
		return false;
	}
	else {
		return true;
	}	
}

// returns true if columns are order with increasing numbers from left to right (e.g. column (0 2 1 0)^T is integer 210)
static inline int is_column_ordered(const int m, const int K, const register int q[]) {
	int i = 0;
	
	if (K==1) {
		return true;
	}
	
	for (i=0; i < (K-1); i++) {
		if (!column_cmp(i+1, m, K, q)) {
			return false;
		}
	}
	
	return true;
}

// returns true if [(1 1)^T | (1 1)^T] or [(1 1)^T | (1 2)^T] are contained in configuration matrix 
// (as in eq. (2.3.7) of T. Huebsch, "Calabi-Yau Manifolds -- A Bestiary for Physicists")
int is_1fold_id(const int m, const int K, const int n_r[], const int q[]) {
	int num_P1 = 0, i = 0, j = 0, l = 0;
	
	while ( (num_P1 < m) && (n_r[num_P1] == 1) ) {
		num_P1++;
	}
	
	if (num_P1 >= 2) {
		for (i=0; i < (num_P1-1); i++) {
			for (j=i+1; j < num_P1; j++) {
				for (l=0; l < K; l++) {
					// also catch [(1 1)^T | (1 2)^T]
					if ( (q[l+(i*K)]>0) && (q[l+(j*K)]>0) && ((q[l+(i*K)]+q[l+(j*K)]) <= 3) && (sum_q_column(l,m,K,q)==(q[l+(i*K)]+q[l+(j*K)])) ) {
						return true;
					}
				}
			}
		}
	}
	
	// new code: check for [2 | 2] subconfig
	if (num_P1 == m) {
		return false;
	}
	int num_P2 = 0;
	while ( ((num_P1+num_P2) < m) && (n_r[num_P1 + num_P2] == 2)) {
		num_P2++;
	}
	if (num_P2 < 1) {
		return false;
	}
	for (i=num_P1; i < (num_P1+num_P2); i++) {
		for (j=0; j < K; j++) {
			if ( (q[j+(i*K)]==2) && (sum_q_column(j,m,K,q)==2) ) {
				return true;
			}
		}
	}
	
	return false;
}

// returns true if q-matrix has a [1 | 2] sub-configuration
int is_0fold_id(const int m, const int K, const int n_r[], const int q[]) {
	int num_P1 = 0, i = 0, j = 0;
	
	while ( (num_P1 < m) && (n_r[num_P1] == 1) ) {
		num_P1++;
	}
	
	if (num_P1 >= 1) {
		for (i=0; i < num_P1; i++) {
			for (j=0; j < K; j++) {
				if ( (q[j+(i*K)]==2) && (sum_q_column(j,m,K,q)==2) ) {
					return true;
				}
			}
		}
	}
	
	return false;
}

// returns true if q-matrix has one of the following sub-configurations:
// [3 | 2], [4 | 2 2], [(1 2)^T | (1 1)^T], [(1 2)^T | (2 1)^T], [(1 2)^T | (1 3)^T], [(1 2)^T | (2 2)^T] or [(1 1 1)^T | (1 1 1)^T]
int is_2fold_id(const int m, const int K, const int n_r[], const int q[]) {
	int i = 0, j = 0, l = 0, p = 0, num_P1 = 0; //first_4_2 = 0;
	
	while ( (num_P1 < m) && (n_r[num_P1] == 1) ) {
		num_P1++;
	}
	
	for (i=num_P1; i < m; i++) {
		// [3 | 2]
		if (n_r[i] == 3) {
			for (j=0; j < K; j++) {
				if ( (q[j+(i*K)]==2) && (sum_q_column(j,m,K,q)==2) ) {
					return true;
				}
			}
		}
		// [4 | 2 2]
//		if (n_r[i] == 4) {
//			first_4_2 = 0;
//			for (j=0; j < K; j++) {
//				if ( (q[j+(i*K)]==2) && (sum_q_column(j,m,K,q)==2) ) {
//					if (first_4_2 == 1) {
//						return true;
//					}
//					first_4_2 = 1;
//				}				
//			}
//		}
//		if (n_r[i] > 4) {
		if (n_r[i] > 3) {
			break;
		}
	}
	
	// [(1 2)^T | (1 1)^T], [(1 2)^T | (2 1)^T], [(1 2)^T | (1 3)^T] and [(1 2)^T | (2 2)^T]
	for (i=0; i < num_P1; i++) {
//		if (n_r[i] != 1) {
//			break;
//		}
		
		for (j=num_P1; j < m; j++) {
			if (n_r[j] != 2) {
				continue;
			}
			
			for (l=0; l < K; l++) {
				if ( ((q[l+(i*K)]==1) && (q[l+(j*K)]==1)) && (sum_q_column(l,m,K,q)==2) ) {
					return true;
				}
				if ( ((q[l+(i*K)]==2) && (q[l+(j*K)]==1)) && (sum_q_column(l,m,K,q)==3) ) {
					return true;
				}
//				if ( ((q[l+(i*K)]==1) && (q[l+(j*K)]==3)) && (sum_q_column(l,m,K,q)==4) ) {
//					return true;
//				}
				if ( ((q[l+(i*K)]==2) && (q[l+(j*K)]==2)) && (sum_q_column(l,m,K,q)==4) ) {
					return true;
				}
			}
		}
	}
	
	// [(1 1 1)^T | (1 1 1)^T], [(1 1 1)^T | (1 1 2)^T] and [(1 1 1)^T | (1 2 2)^T]
	if (num_P1 >= 3) {
		for (i=0; i < (num_P1-2); i++) {
			for (j=i+1; j < (num_P1-1); j++) {
				for (l=j+1; l < (num_P1); l++) {
					for (p=0; p < K; p++) {
						// [(1 1 1)^T | (1 1 1)^T]
						if ( ((q[p+(i*K)]==1) && (q[p+(j*K)]==1) && (q[p+(l*K)]==1)) && (sum_q_column(p,m,K,q)==3) ) {
							return true;
						}
						// [(1 1 1)^T | (1 1 2)^T]
						if ( ((q[p+(i*K)]==1) && (q[p+(j*K)]==1) && (q[p+(l*K)]==2)) && (sum_q_column(p,m,K,q)==4) ) {
							return true;
						}
						if ( ((q[p+(i*K)]==1) && (q[p+(j*K)]==2) && (q[p+(l*K)]==1)) && (sum_q_column(p,m,K,q)==4) ) {
							return true;
						}
						if ( ((q[p+(i*K)]==2) && (q[p+(j*K)]==1) && (q[p+(l*K)]==1)) && (sum_q_column(p,m,K,q)==4) ) {
							return true;
						}
						// [(1 1 1)^T | (1 2 2)^T]
						if ( ((q[p+(i*K)]==1) && (q[p+(j*K)]==2) && (q[p+(l*K)]==2)) && (sum_q_column(p,m,K,q)==5) ) {
							return true;
						}
						if ( ((q[p+(i*K)]==2) && (q[p+(j*K)]==1) && (q[p+(l*K)]==2)) && (sum_q_column(p,m,K,q)==5) ) {
							return true;
						}
						if ( ((q[p+(i*K)]==2) && (q[p+(j*K)]==2) && (q[p+(l*K)]==1)) && (sum_q_column(p,m,K,q)==5) ) {
							return true;
						}
					}
				}
			}
		}
	}
	
	return false;
}

// compute x^y and returns result (for x,y positive integers)
static inline int mypow(const register int x, register int y) {
	int ret = 1;
	
	while (y--) {
		ret *= x;
	}
	return ret;
}

static inline void next_idx1(const int m, const int K, register int idx[]) {
	int i=K-1;
	
	while ( (i >= 0) && (idx[i] == (m-1)) ) {
		i--;
	}
	
	if (i < 0) {
		return;
	}
	
	idx[i]++;
	if (i < (K-1)) {
		memset(&idx[i+1], 0, sizeof(int)*(K-i-1));
	}
}

static inline int is_nilpotent(const int m, const int K, const int n_r[], const int idx[]) {
	int i;
	int power[m];
	memset(power, 0, sizeof(int)*m);
	
	for (i=0; i < K; i++) {
		power[idx[i]]++;
	}
	for (i=0; i < m; i++) {
		if (power[i] > n_r[i]) {
			return true;
		}
	}
	
	return false;
}

void compute_indices_3fold(int indices[4], const int eulernumber, const int m, const int K, const int n_r[], const int q[]) {
	float A=0, B=0, C=0, E=0;
	float A_tmp, B_tmp, C_tmp, E_tmp;
	int i, j, l, r;// tmp;
	int power[m], bNilpotent;

	unsigned long long len = mypow(m, K), p;
	int idx[K], val;
	
	for (i = 0; i < m; i++) {
		for (j = 0; j < m; j++) {
			for (l = 0; l < m; l++) {
				// A = \int_M c_2(M) c_1(N); store A^{ijl} in A
				// B = \int_M ch_3(N); store B^{ijl} in B
				// C = \int_M c_1(N)^3; store C^{ijl} in C
				// E = \int_M c_1(N) ch_2(N); store E^{ijl} in E
				A_tmp = (i == j) ? -(n_r[i] + 1) : 0; B_tmp = 0; C_tmp = 0; E_tmp = 0;
				for (p = 0; p < K; p++) {
					A_tmp += q[p+(i*K)] * q[p+(j*K)];
					B_tmp += q[p+(i*K)] * q[p+(j*K)] * q[p+(l*K)];
					E_tmp += q[p+(i*K)] * q[p+(j*K)];
				}
				A_tmp *= n_r[l] + 1;
				A_tmp *= 0.5;
				B_tmp /= 6;
				C_tmp = (n_r[i] + 1) * (n_r[j] + 1) * (n_r[l] + 1);
				E_tmp *= n_r[l] + 1;
				E_tmp *= 0.5;

				// the following if-test is almost never true, thus: not needed
				/*if ( (A_tmp == 0) && (B_tmp == 0)  && (C_tmp == 0) && (E_tmp == 0) ) {
					continue;
				}*/

				memset(idx, 0, sizeof(int)*K);	// no difference if '*' in last arg is stripped out of for-loop by introd. of new var.
				for (p = 0; p < len; p++) {
					if (p > 0) {
						next_idx1(m, K, idx);
					}
					val = 1;
					
					bNilpotent = false;
					memset(power, 0, sizeof(int)*m);	// no difference if '*' in last arg is stripped out of for-loop by introd. of new var.
					for (r=0; r < K; r++) {
						val *= q[r+((idx[r])*K)];
						power[idx[r]]++;
					}
					power[i]++; power[j]++; power[l]++;
					for (r=0; r < m; r++) {		// for (r = m; r--; ) { // <-- not faster
						if (power[r] != n_r[r]) {
							bNilpotent = true;
							break;
						}
					}
					if (!bNilpotent) {
						A += A_tmp * val;
						B += B_tmp * val;
						C += C_tmp * val;
						E += E_tmp * val;
					}
				}
			}
		}
	}

	indices[0] = eulernumber * 0.5;	// case (k,l)=(0,1) is just half the eulernumber
	indices[1] = lround((A + (12*B))/12);	// case (k,l)=(1,0) is equal to \chi^h(N)
	indices[2] = lround((2*K*indices[1]) + E + E);
	indices[3] = lround((3*K*K*indices[1]) + C + (6*K*E));
	// in above expressions: lround() only needed to get e.g. 125(int) for 124.999...(float) <== how 125 is stored after above calculation

// NOTE: possible float --> int conversion errors! (can use the syntax %.0f below to convert safely via sprintf and atoi)
//	printf("(%i,%i): %.0f [%i] ", i, j, tmp, (int)tmp);
//	printf("(1,0): %f, (2,0): %f, (3,0): %f\n", (A + (12*B))/12, (2*K*(A + (12*B))/12) + (2*E), (3*K*K*(A + (12*B))/12) + C + (6*K*E));
// debug:
//	indices[0] = A;
//	indices[1] = B;
//	indices[2] = C;
//	indices[3] = E;
//	printf("\n");
//	printf("A = %f, B = %f, C = %f, E = %f\n", A, B, C, E);
}

// (note: ~18% performance boost here with C++ stl::set as compared to old hand-made c linked list)
inline int is_already_there(set<qnode> * matrixlist, int m, int K, int n_r[], int q[], int num_perms, int *n_r_perms[]) {
	int i, j;
	qnode perm_qnode;
	int permuted_q[m*K];

	perm_qnode.len = m*K;
	for (i = 0; i < num_perms; i++) {
		// generate permuted q-matrix
		for (j = 0; j < m; j++) {
			mymemcpy(&permuted_q[j*K], &q[(n_r_perms[i][j])*K], K);
		}
		
		// sort columns of permuted q-matrix
		sort_q(m, K, permuted_q);
		
		// walk the q-matrix set
		perm_qnode.q = permuted_q;
		if (matrixlist->find(perm_qnode) != matrixlist->end()) {
			return true;
		}
	}

	return false;
}

inline void sort_q(const int m, const int K, int q[]) {
	int i, j;
	int columns[K][m];
	
	// decompose q into K column vectors (K 1-d arrays)
	for (i = 0; i < K; i++) {
		for (j = 0; j < m; j++) {
			columns[i][j] = q[i + (j*K)];
		}
	}
	
	// sort using qsort library function
	m_tmp = m;	// need this to tell col_cmp function how long the column vectors are (i.e. length: m)
	qsort(columns, K, m*sizeof(int), col_cmp);
	
	// rebuild q-matrix out of the ordered column vectors
	for (i = 0; i < K; i++) {
		for (j = 0; j < m; j++) {
			q[i + (j*K)] = columns[i][j];
		}
	}
}

// qsort column array comparison function
int col_cmp(const void *a, const void *b) {
	int i = 0, ret = 0;
	
	const int * ia = (const int *)a; // casting pointer types
	const int * ib = (const int *)b;
	
	while ((i < m_tmp) && ((ret = ia[i] - ib[i]) == 0)) {
		i++;
	}
	
	return ret; // integer comparison: returns negative if b > a and positive if a > b
}

inline void init_num_P(const int m, const int n_r[], int num_P[]) {
	int i;
	
	for (i = 0; i < m; i++) {
		num_P[n_r[i]-1]++;
	}
}

inline int calc_num_perms(const int num_P[], const int max_P) {
	int i, j;
	int tmp = 1, ret = 1;
	
	for (i = 0; i < max_P; i++) {
		tmp = 1;
		// this for-loop computes the factorial of num_P[i], i.e. (num_P[i])!
		for (j = 1; j <= num_P[i]; j++) {
			tmp *= j;
		}
		ret *= tmp;
	}
	
	return ret;
}

int** init_permlist(const int m, const int n_r[], const int num_P[], const int num_perms) {
	int i;
	int **n_r_perms;

	n_r_perms = (int **)malloc(num_perms*sizeof(int*));
	
	for (i = 0; i < num_perms; i++) {
		n_r_perms[i] = (int *)malloc(m*sizeof(int));
	}

	permlistcounter = 0; 
	tmpperm = (int*)malloc(m*sizeof(int));
	for (i = 0; i < m; i++) {
		tmpperm[i] = i;	// set start value: 1st permutation is trivial (0,1,2,...,m-1)
	}
	
	permutation(m, n_r, num_P, num_perms, n_r_perms, 0);

	free(tmpperm); tmpperm = NULL;
	
	// debug
/*	int j;
	printf("(%i", n_r[0]);
	for (i = 1; i < m; i++) {
		printf(",%i", n_r[i]);
	}
	printf("):\n");
	for (i = 0; i < num_perms; i++) {
		printf("(%i", n_r_perms[i][0]);
		for (j = 1; j < m; j++) {
			printf(",%i", n_r_perms[i][j]);
		}
		printf(")\n");
	}
	printf("\n");
*/	
	return n_r_perms;
}

// shifts all elements in "perm" one position to the left
void perm_rotate(int perm[], const int count) {
	int i;
    int tmp = perm[0];
	
    for (i = 0; i < count - 1; i++) {
		perm[i] = perm[i + 1];
	}
    perm[count - 1] = tmp;
}

void permutation(const int m, const int n_r[], const int num_P[], const int num_perms, int *n_r_perms[], int val) {
	int i;
	
    if (m == val) {
		// we've produced a new permutation, but we need to check that it is a valid one, i.e. rows are only
		// swapped if the dimension of the respective projective spaces is equal
		for (i = 0; i < m; i++) {
			if (n_r[tmpperm[i]] != n_r[i]) {
				return;
			}
		}
		
		// debug
//		printf("(%i", tmpperm[0]);
//		for (i = 1; i < m; i++) {
//			printf(",%i", tmpperm[i]);
//		}
//		printf(")\n");
		
		mymemcpy(n_r_perms[permlistcounter++], tmpperm, m);
	}
    else {
		for (i = m - val; i > 0 ; i--) {
			permutation(m, n_r, num_P, num_perms, n_r_perms, val+1);
			perm_rotate(tmpperm + val, m - val);
		}
    }
}

inline void free_permlist(const int num_perms, int *n_r_perms[]) {
	int i;

	for (i = 0; i < num_perms; i++) {
		free(n_r_perms[i]);
	}
}