/*
 *  is_already_there.h
 *  The function "is_already_there" has 2 arguments: a list "conflist" of matrices and a candidate configuration [n|q].
 *  It creates all permutations of [n|q] and checks if any of them is contained in "conflist". 
 *  If it finds a match the position in conflist is returned, otherwise conflist->end() is returned.
 *  In our paper ITP-UH-04/13, this is referred to as the "brute force" permution comparison algorithm.
 *
 *  Created by Alexander Haupt on 11/06/2012.
 *  Copyright 2013 Alexander Haupt. All rights reserved.
 *
 */

#if defined(_WIN32) || defined(_WIN64)
  #error not tested on windows
#endif

#include <string.h>
#include <set>
#include <algorithm>
#include <stdio.h>

#include "mymemcpy.h"

using namespace std;
using namespace __gnu_cxx;

extern int m_tmp;
//int test_permcounter;	// debug

struct conf {
	int m;
	int K;
	int * n_r;
	int * q;
	int eulernumber;
};

int col_cmp(const void *a, const void *b);
bool operator <(conf a, conf b);
//inline void print_perm(const int m, int owntmpperm[]);	// debug

inline void sort_q(const int m, const int K, int q[]) {
	int i, j;
	int columns[K][m];
	
	// decompose q into K column vectors (K 1-d arrays)
	for (i = 0; i < K; i++) {
		for (j = 0; j < m; j++) {
			columns[i][j] = q[i + (j*K)];
		}
	}
	
	// sort using qsort library function
	m_tmp = m;	// need this to tell col_cmp function how long the column vectors are (i.e. length: m)
	qsort(columns, K, m*sizeof(int), col_cmp);
	
	// rebuild q-matrix out of the ordered column vectors
	for (i = 0; i < K; i++) {
		for (j = 0; j < m; j++) {
			q[i + (j*K)] = columns[i][j];
		}
	}
};

inline set<conf>::iterator is_already_there_helper(set<conf> * conflist, const int m, const int K, int n_r[], int q[], const int from, const int to, const int n_r_pos, const int num_diffpns, int pn_range[], int owntmpperm[]) {
	int i;
	conf perm_conf;
	int permuted_q[m*K];
	set<conf>::iterator it;
	
	perm_conf.m = m;
	perm_conf.K = K;
	perm_conf.n_r = n_r;
	//printf("%i, %i\n", from, to);	// debug
	
	do {
		//print_perm(m, owntmpperm);	// debug

		// do we need to permute further?
		if (n_r_pos < num_diffpns) {
			it = is_already_there_helper(conflist, m, K, n_r, q, to, to+pn_range[n_r_pos], n_r_pos+1, num_diffpns, pn_range, owntmpperm);
			if (it != conflist->end()) {
				return it;
			}
		}
		else {
			//printf("\r%i", ++test_permcounter);	// debug
			//++test_permcounter;	// debug
			
			// generate permuted q-matrix
			for (i = 0; i < m; i++) {
				mymemcpy(&permuted_q[i*K], &q[(owntmpperm[i])*K], K);
			}

			// sort columns of permuted q-matrix
			sort_q(m, K, permuted_q);

			// walk the q-matrix set
			perm_conf.q = permuted_q;
			if ( (it = conflist->find(perm_conf)) != conflist->end() ) {
				return it;
			}
		}
	} while ( next_permutation (owntmpperm+from,owntmpperm+to) );
	
	return conflist->end();
};

// (note: ~18% performance boost here with C++ stl::set as compared to old hand-made c linked list)
inline set<conf>::iterator is_already_there(set<conf> * conflist, const int m, const int K, int n_r[], int q[]) {
	int i;
	set<conf>::iterator it;

	int num_diffpns = 1;
	int pn_range[10] = {0,0,0,0,0,0,0,0,0,0}; // CAREFUL for n = 5 (i.e. n > 4)

	int * owntmpperm = new int[m];
	for (i = 0; i < m; i++) {
		owntmpperm[i] = i;	// set start value: 1st permutation is trivial (0,1,2,...,m-1)

		if ( (i>0) && (n_r[i-1] != n_r[i]) ) {
			num_diffpns++;
		}
		pn_range[num_diffpns-1]++;
	}
	// debug
	//printf("%i, (%i,%i,%i,%i,%i,%i,%i,%i,%i,%i)\n", num_diffpns, pn_range[0], pn_range[1], pn_range[2], pn_range[3], pn_range[4], pn_range[5], pn_range[6], pn_range[7], pn_range[8], pn_range[9]); fflush(stdout);

	//test_permcounter = 0;	// debug
	it = is_already_there_helper(conflist, m, K, n_r, q, 0, pn_range[0], 1, num_diffpns, pn_range, owntmpperm);
	//printf("%i", test_permcounter);	// debug
	// debug
	//for (i = 0; i < m; i++) {
	//	printf("%i ", owntmpperm[i]);
	//}
	//printf("\n");
	// end: debug
	
	delete[] owntmpperm;
	return it;
};

inline void add_conf(set<conf> * conflist, const int m, const int K, const int n_r[], const int q[], const int eulernumber) {
	conf new_conf;
	
	new_conf.m = m;
	new_conf.K = K;
	new_conf.eulernumber = eulernumber;

	new_conf.n_r = (int*) malloc( sizeof(int)*m );	
	memcpy(new_conf.n_r, n_r, sizeof(int)*m);	
	
	new_conf.q = (int*) malloc( sizeof(int)*(m*K) );	
	memcpy(new_conf.q, q, sizeof(int)*(m*K));
	
	conflist->insert(new_conf);
};