/*
 *  mu.h
 *  Code to compute the embedding form \mu, to check the \mu-criterion and to compute the 
 *  Euler characteristic of zero and one dimensional singular loci of determinantal varieties.
 *
 *  Created by Alexander Haupt on 29/02/2012.
 *  Copyright 2013 Alexander Haupt. All rights reserved.
 *
 */

#if defined(_WIN32) || defined(_WIN64)
  #error not tested on windows
#endif

#include <stdio.h>
#include <vector>

#include "mymemcpy.h"

using namespace std;
using namespace __gnu_cxx;

typedef struct term {
	int coeff;
	int * Jpower;
} jmonomial;

// debug
inline void print_mu(const int m, vector<jmonomial> * mu) {
	int i, j;
  	
	printf("mu = ");
	for (i = 0; i < mu->size(); i++) {
		printf("+ %i ", mu->at(i).coeff);
		for (j = 0; j < m; j++) {
			if (mu->at(i).Jpower[j] != 0) {
				printf("J[%i]^%i ", j+1, mu->at(i).Jpower[j]);
			}
		}
	}
	printf("\n");
};

// free everything to avoid memory leaks
inline void clear_mu(vector<jmonomial> * mu) {
	int i;
	for (i = 0; i < mu->size(); i++) {
		delete[] mu->at(i).Jpower;
	}
	
	mu->clear();
};

inline void myerase (vector<jmonomial> * mu, const int i) {
	delete[] mu->at(i).Jpower;
	mu->erase(mu->begin()+i);
};

inline void myerase (vector<jmonomial> * mu, const int start, const int end) {
	int i;
	for (i = start; i < end; i++) {
		delete[] mu->at(i).Jpower;
	}
	mu->erase(mu->begin()+start, mu->begin()+end);
};

// simplify (group together equal terms)
inline void jmonomial_simplify(const int m, vector<jmonomial> * mu) {
	int groupterms, j, l, l1;

	for (j = 0; j < mu->size(); j++) {
		for (l = j+1; l < mu->size(); l++) {
			groupterms = 1;
			for (l1 = 0; l1 < m; l1++) {
				if (mu->at(j).Jpower[l1] != mu->at(l).Jpower[l1]) {
					groupterms = 0;
					break;
				}
			}
			if (groupterms == 1) {
				mu->at(j).coeff += mu->at(l).coeff;
				myerase(mu, l);
				l--;
			}
		}
	}
};

inline void calc_mu(const int m, const int K, const register int n_r[], const register int q[], vector<jmonomial> * mu) {
	int i, j, l, max;
	jmonomial * tmpterm;
	
	//debug_output(m, K, n_r, q); // debug
	
	// initialize mu from first column in q
	for (i = 0; i < m; i++) {
		if (q[i*K] == 0) {continue;}
		
		tmpterm = new jmonomial;
		tmpterm->coeff = q[i*K];
		tmpterm->Jpower = new int[m];
		for (j = 0; j < m; j++) {
			tmpterm->Jpower[j] = (i==j) ? 1 : 0;
		}
		mu->push_back(*tmpterm);
		delete tmpterm;
	}

	for (i = 1; i < K; i++) {
		max = mu->size();
		for (l = 0; l < max; l++) {
			for (j = 0; j < m; j++) {
				if ((mu->at(l).Jpower[j] < n_r[j]) && (q[i+(j*K)] != 0)) {
					tmpterm = new jmonomial;
					tmpterm->coeff = mu->at(l).coeff * q[i+(j*K)];
					tmpterm->Jpower = new int[m];
					mymemcpy(tmpterm->Jpower, mu->at(l).Jpower, m);
					tmpterm->Jpower[j]++;
					mu->push_back(*tmpterm);
					delete tmpterm;
				}
			}
		}
		myerase(mu, 0, max);
		//print_mu(m, mu); // debug
		
		if (mu->size() == 0) {break;} // \mu == 0
		  
		// simplify (group together equal terms)
		int groupterms, l1;
		for (j = 0; j < mu->size(); j++) {
			for (l = j+1; l < mu->size(); l++) {
				groupterms = 1;
				for (l1 = 0; l1 < m; l1++) {
					if (mu->at(j).Jpower[l1] != mu->at(l).Jpower[l1]) {
						groupterms = 0;
						break;
					}
				}
				if (groupterms == 1) {
					mu->at(j).coeff += mu->at(l).coeff;
					myerase(mu, l);
					l--;
				}
			}
		}
		//print_mu(m, mu); // debug
	}
	//print_mu(m, mu); // debug
};

inline int compute_detvar_eulernumber_fast(const int dim, const int m, const int K, const register int n_r[], const register int q[], vector<jmonomial> * mu) {
	int i, j, topform, chi;
	
	if (mu->empty()) {return 0;} // \mu == 0
	
	chi = 0;
	if (dim == 0) {
		//debug_output(m, K, n_r, q); // debug
		for (i = 0; i < mu->size(); i++) {
			topform = 1;
			for (j = 0; j < m; j++) {
				if (mu->at(i).Jpower[j] != n_r[j]) {
					topform = 0;
					break;
				}
			}
			if (topform == 1) {
				chi += mu->at(i).coeff;
			}
		}
		//print_mu(m, mu); // debug
		//printf("chi = %i\n", chi); // debug
	}
	else if (dim == 1) {
		int l;
		int c1_coeff[m]; 
		//debug_output(m, K, n_r, q); // debug
		//printf("c1 = "); // debug
		for (i = 0; i < m; i++) {
			c1_coeff[i] = n_r[i] + 1;
			for (j = 0; j < K; j++) {
				c1_coeff[i] -= q[j + (i*K)];
			}
			//printf(" + %i J[%i]", c1_coeff[i], i+1); // debug
			
			if (c1_coeff[i] == 0) {continue;}
			
			for (j = 0; j < mu->size(); j++) {
				topform = 1;
				for (l = 0; l < m; l++) {
					if (mu->at(j).Jpower[l] != n_r[l] - ((i==l) ? 1 : 0)) {
						topform = 0;
						break;
					}
				}
				if (topform == 1) {
					chi += mu->at(j).coeff * c1_coeff[i];
				}
			}
		}
		//printf("\n"); // debug
		//print_mu(m, mu); // debug
		//printf("chi = %i\n", chi); // debug
	}
	else {
		printf("Function \"compute_detvar_eulernumber_fast\" only implemented for dim=0,1.\n");
	}
	
	return chi;
};

// returns TRUE if detvar (here: in q-array) represents an empty set (for n==4, i.e. CICY 4-folds !)
inline int is_detvar_emptyset(const int m, const int K, const register int n_r[], const register int q[], vector<jmonomial> * mu) {
	int i, j, l, topform;
	int chixprime = 0;

	//debug_output(m, K, n_r, q); // debug
	//print_mu(m, mu); // debug
	
	if (mu->size() == 0) {return true;} // \mu == 0
	  
	// compute \chi(X'), see paper ITP-UH-04/13
	for (i = 0; i < m; i++) {
		for (j = 0; j < mu->size(); j++) {
			topform = 1;
			for (l = 0; l < m; l++) {
				if (mu->at(j).Jpower[l] != n_r[l] - ((i==l) ? 1 : 0)) {
					topform = 0;
					break;
				}
			}
			if (topform == 1) {
				chixprime += mu->at(j).coeff;
			}
		}
	}
	//printf("chi(X') = %i\n", chixprime); // debug
	
	if (chixprime == 0) {
		return true;
	}
	else {
		return false;
	}
};