/*
 *  qmat_cls.cpp
 *  q-matrix class member functions. This class is only used in 2ndfilter_2.cpp
 *
 *  Created by Alexander Haupt on 09/03/2011.
 *  Copyright 2013 Alexander Haupt. All rights reserved.
 *
 */

#include "qmat_cls.h"

void qmat::setq (const int new_m, const int new_K, const int new_q[]) {
	int i, j;
	
	if ( (new_m == 0) || (new_K == 0) || (new_q == NULL) ) {
		return;
	}
	
	m = new_m;
	K = new_K;
	
	q = new int[m*K];
	memcpy(q, new_q, sizeof(int)*m*K);
	
	rows = new int* [m];
	cols = new int* [K];

	// decompose q into m row vectors (m 1d arrays of length K)
	for (i = 0; i < m; i++) {
		rows[i] = new int[K]; //(int *)malloc(K*sizeof(int));
		for (j = 0; j < K; j++) {
			rows[i][j] = q[j + (i*K)];
		}
	}

	// decompose q into K column vectors (K 1d arrays of length m)
	for (i = 0; i < K; i++) {
		cols[i] = new int[m]; //(int *)malloc(m*sizeof(int));
		for (j = 0; j < m; j++) {
			cols[i][j] = q[i + (j*K)];
		}
	}
}

void qmat::clear () {
	int i;
	for (i = 0; i < m; i++) {
		delete rows[i];
	}
	for (i = 0; i < K; i++) {
		delete cols[i];
	}
	delete rows; delete cols;
	
	m = 0;
	K = 0;
	delete q;
	q = NULL;
	rows = NULL;
	cols = NULL;	
}

qmat::qmat () {
	m = 0;
	K = 0;
	q = NULL;
	rows = NULL;
	cols = NULL;
}

qmat::qmat (const int new_m, const int new_K, const int new_q[]) {
	setq(new_m, new_K, new_q);
}

qmat::~qmat () {
	clear();
}

int * qmat::delete_row(const int row) {
	int i, j;
	int * new_q = new int[(m-1)*K];
	
	for (i = 0; i < m; i++) {
		if (i < row) {
			for (j = 0; j < K; j++) {
				new_q[j + (i*K)] = rows[i][j];
			}
		}
		if (i > row) {
			for (j = 0; j < K; j++) {
				new_q[j + ((i-1)*K)] = rows[i][j];
			}
		}
	}
	
	return new_q;	
}

int * qmat::delete_col(const int col) {
	int i, j;
	int * new_q = new int[m*(K-1)];
	
	for (i = 0; i < K; i++) {
		for (j = 0; j < m; j++) {
			if (i < col) {
				new_q[i + (j*(K-1))] = cols[i][j];
			}
			if (i > col) {
				new_q[(i-1) + (j*(K-1))] = cols[i][j];
			}
		}
	}
	
	return new_q;
}

int * qmat::row(const int idx) {
	int i;
	int * new_row = new int[K];
	
	for (i = 0; i < K; i++) {
		new_row[i] = rows[idx][i];
	}
	
	return new_row;
}

int * qmat::col(const int idx) {
	int i;
	int * new_col = new int[m];
	
	for (i = 0; i < m; i++) {
		new_col[i] = cols[idx][i];
	}
	
	return new_col;
}

int * qmat::add_row(const int row[]) {
	int i, j;
	int * new_q = new int[(m+1)*K];
	
	memcpy(new_q, row, sizeof(int)*K);
	for (i = 0; i < m; i++) {
		for (j = 0;	j < K; j++) {
			new_q[j + ((i+1)*K)] = rows[i][j];
		}
	}
	
	return new_q;
}

int * qmat::add_col(const int col[]) {
	int i, j;
	int * new_q = new int[m*(K+1)];
	
	for (i = 0; i < m; i++) {
		for (j = 0;	j < K; j++) {
			new_q[j + (i*K)] = cols[j][i];
		}
		new_q[K + (i*K)] = col[i];
	}
	
	return new_q;
}

// adds two rows together and returns the result
int * qmat::row_addition(const int a, const int b) {
	int i;
	int * new_row = new int[K];
	
	for (i = 0; i < K; i++) {
		new_row[i] = rows[a][i] + rows[b][i];
	}
	
	return new_row;
}

// multiply each element of row by val and return new full qmatrix
int * qmat::row_mult(const int idx, const int val) {
	int i;
	int * new_q = new int[m*K];
	
	//new_q = (int*) malloc( sizeof(int)*m*K );
	memcpy(new_q, q, sizeof(int)*m*K);
	
	for (i = 0; i < K; i++) {
		new_q[i+(idx*K)] *= val;
	}
	
	return new_q;
}

int * qmat::return_qcopy() {
	int * new_q = new int[m*K];
	
	memcpy(new_q, q, sizeof(int)*m*K);
	
	return new_q;
}

// swap rows a and b and return new full qmatrix
int * qmat::swap_rows(const int a, const int b) {
	int i, j;
	int * new_q = new int[m*K];
	
	for (i = 0; i < m; i++) {
		if (i == a) {
			for (j = 0;	j < K; j++) {
				new_q[j + (b*K)] = rows[i][j];
			}			
		} 
		if (i == b) {
			for (j = 0;	j < K; j++) {
				new_q[j + (a*K)] = rows[i][j];
			}
		}
		if ( (i != a) && (i != b) ) {
			for (j = 0;	j < K; j++) {
				new_q[j + (i*K)] = rows[i][j];
			}
		}
	}
	
	return new_q;
}

// move row at position idx to the top (i.e. as first row) and return new full qmatrix
int * qmat::move_row_to_top(const int idx) {
	int i, j;
	int * new_q = new int[m*K];
	
	memcpy(new_q, rows[idx], sizeof(int)*K);
	
	for (i = 0; i < m; i++) {
		if (i < idx) {
			for (j = 0;	j < K; j++) {
				new_q[j + ((i+1)*K)] = rows[i][j];
			}			
		} 
		if (i > idx) {
			for (j = 0;	j < K; j++) {
				new_q[j + (i*K)] = rows[i][j];
			}
		}
	}
	
	return new_q;
}


// replace all occurrences of val1 by val2 in column idx and return new full qmatrix
int * qmat::replace_in_col(const int idx, const int val1, const int val2) {
	int i;
	int * new_q = new int[m*K];

	memcpy(new_q, q, sizeof(int)*m*K);
	
	for (i = 0; i < m; i++) {
		if (new_q[idx + (i*K)] == val1) {
			new_q[idx + (i*K)] = val2;
		}
	}
	
	return new_q;
}