/*
 *  This program reads a CICY list and performs all possible P1 splittings on the matrices.
 *  It also implements RemoveRedundancy (i.e. the "is_already_there" function) and removes "some" (only between the split matrices) ineffective splittings.
 *
 *  Created by Alexander Haupt on 12/05/2010.
 *  Copyright 2013 Alexander Haupt. All rights reserved.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <time.h>
#include <set>
#include <vector>
#include <algorithm>
#include <sstream>
#include <string>
#include <sys/stat.h>
#include <errno.h>

#if defined(__APPLE__) || defined(__MACH__) || defined(__OSX__)
  #include <malloc/malloc.h>
#else
  #include <malloc.h>
#endif

#include "is_already_there.h"
#include "mu.h"

using namespace std;
using namespace __gnu_cxx;


int n;
int new_counter;

void splitting(const int num, const char filename[]);
static inline void write_output1(FILE * fp, const int counter, const int m, const int K, const register int n_r[], const register int q[], const int eulernumber);
inline void debug_output(const int m, const int K, const int n_r[], const int q[]);
inline int * init_array(const int dim, const int val);
inline void init_val(const int dim, int array[], const int val);
inline int sum_q_column(const int a, const int m, const int K, const int q[]);
void split_matrix(FILE * fp, const conf conflist[], const int idx, set<conf> * matrixlist);
inline int cols_equal(const int m, const int K, const int q[], const int col1, const int col2);
int increase_subsplit(const int m, register int q[]);
inline int is_equal_ambient_space(const int m, const int n_r1[], const int n_r2[]);
inline int is_triv_subsplit(const int m, const int subsplit[]);


int main (int argc, const char * argv[]) {
	if (argc != 4) {
		// input: "n" -- for CICY n-fold; "num" -- number of configuration matrices in file
		printf("Usage: %s n num filename\n", argv[0]);
	}
	else {
		n = atoi(argv[1]);
		splitting(atoi(argv[2]), argv[3]);
	}
	
	return 0;
}

void splitting(const int num, const char filename[]) {
	int i, linecounter = 0, counter = 0;
	int m, K, eulernumber, orig_matno;
	int * n_r;
	int * q;	// q part of CICY config. matrix

	FILE * fp = fopen(filename, "r");

	conf conflist[num];
	set<conf> matrixlist;
	set<conf> final_matrixlist;
	set<conf>::iterator it;
	
	char output_filename[50]; time_t timer;
	timer = time(NULL);
	sprintf(output_filename, "splitting_%s", basename(filename));
	FILE * fp1 = fopen(output_filename, "a");	
	
	if (fp != NULL) {
		char line[128], strnum[2] = {0,0};
		
		while(fgets(line, sizeof line, fp) != NULL) { /* read a line */
			switch (linecounter) {
				case 0:
					if (line[0] == 'M') {
						sscanf(line, "MATRIX NUMBER :%6i\n", &orig_matno);
						linecounter++;
					}
					break;
				case 1:
					sscanf(line, "SIZE :%4i%3i\n", &m, &K);
					n_r = (int*)malloc(m*sizeof(int));
					q = (int*)malloc(m*K*sizeof(int));
					linecounter++;
					//printf("m=%i, K=%i, ", m, K);	// debug
					break;
				case 2:
					sscanf(line, "EULERNUMBER :%5i\n", &eulernumber);
					linecounter++;
					//printf("eulernumber=%i\n", eulernumber);	// debug
					break;
				default:
					if (linecounter <= (m+2)) {
						n_r[linecounter-3] = 0;
						for (i = 0; i < K; i++) {
							strnum[0] = line[(2*i)+4];
							q[i+((linecounter-3)*K)] = atoi(strnum);
							n_r[linecounter-3] += q[i+((linecounter-3)*K)];
						}
						n_r[linecounter-3]--;
						// only keep CICYs which are embedded in ambient spaces that don't contain P1s
						/*if ( (linecounter == 3) && (n_r[linecounter-3] == 1) ) {
							linecounter = 0;
							continue;
						}*/
						if (linecounter == (m+2)) {
							conflist[counter].m = m;
							conflist[counter].K = K;
							conflist[counter].n_r = n_r;
							conflist[counter].q = q;
							conflist[counter].eulernumber = eulernumber;
							counter++;
						}
						linecounter++;
					}
					else {
						linecounter = 0;
					}
					break;
			}
		}
		if (ferror(fp)) {fprintf(stderr, "File read error. %i: %s\n", ferror(fp), strerror(ferror(fp))); return;}
		fclose(fp);
	}
	else {
		perror(filename); /* why didn't the file open? */
	}
	
	printf("%i %i\n", num, counter);
	new_counter = 0;
	for (i = 0; i < num; i++) {
		time(&timer); printf("%i:\t%s", i+1, ctime(&timer)); fflush(stdout);
		if ( (i > 0) && !((conflist[i-1].m == conflist[i].m) && (is_equal_ambient_space(conflist[i-1].m, conflist[i-1].n_r, conflist[i].n_r) == true)) ) {
			// ^^ that means we are changing to a new ambient space and we must clear the list of matrices with the same ambient space
			copy(matrixlist.begin(), matrixlist.end(), inserter( final_matrixlist, final_matrixlist.end() ) );
			matrixlist.clear();
		}
		split_matrix(fp1, conflist, i, &matrixlist);
	}

	copy(matrixlist.begin(), matrixlist.end(), inserter( final_matrixlist, final_matrixlist.end() ) );
	matrixlist.clear();
	fclose(fp1);

	// write final_matrixlist
	char output_filename2[50];
	sprintf(output_filename2, "splitting_final_%s", basename(filename));
	FILE * fp2 = fopen(output_filename2, "a");
	new_counter = 0;
	for ( it=final_matrixlist.begin(); it != final_matrixlist.end(); it++ ) {
		write_output1(fp2, ++new_counter, (*it).m, (*it).K, (*it).n_r, (*it).q, (*it).eulernumber);
	}
	final_matrixlist.clear();
	fclose(fp2);
	
	// checking that writing of final output file was succesful otherwise write to stderr
	int fs_err = 0;
	struct stat st;
	if (stat(output_filename2, &st) != 0) {fprintf(stderr, "stat() error, errno: %i\n", errno); fs_err = 1;}
	unsigned int size = (unsigned int)st.st_size;
	printf("Output file size: %u\n", size);
	if ( (size == 0) || (fs_err == 1) ) {
		new_counter = 0;
		for ( it=final_matrixlist.begin(); it != final_matrixlist.end(); it++ ) {
			write_output1(stderr, ++new_counter, (*it).m, (*it).K, (*it).n_r, (*it).q, (*it).eulernumber);
		}
	}
}

static inline void write_output1(FILE * fp, const int counter, const int m, const int K, const register int n_r[], const register int q[], const int eulernumber) {
	int i, j;
	char buf[200];
	char * cstr;
	std::stringstream sstm;
	std::string result;
	
	if (sprintf(buf, "MATRIX NUMBER :%6i\nSIZE :%4i%3i\nEULERNUMBER :%5i\n", counter, m, K, eulernumber) < 0) { fprintf(stderr, "An sprintf error occured in write_output1 (1st sprintf)\n"); }
	sstm << buf;
	for (i=0; i < m; i++) {
		sstm << n_r[i] << " |";
		for (j=0; j < K; j++) {
			sstm << " " << q[j+(i*K)];
		}
		sstm << "\n";
	}
	sstm << "\n";
	
	result = sstm.str();
	cstr = new char[result.size()+1];
	strcpy(cstr, result.c_str());
	while (fprintf(fp, "%s", cstr) < 0) {fprintf(stderr, "An fprintf error occured (now try again till write success) for string:\n%s", cstr);}
	fflush(fp);
	delete[] cstr;
}

inline void debug_output(const int m, const int K, const int n_r[], const int q[]) {
	int i, j;
	
	for (i = 0; i < m; i++) {
		printf("%i |", n_r[i]);
		for (j = 0; j < K; j++) {
			printf(" %i", q[j+(i*K)]);
		}
		printf("\n");
	}
	printf("\n");
}


inline int * init_array(const int dim, const int val) {
	int * pointer;
	pointer = (int*)malloc(dim*sizeof(int));
	
	init_val(dim, pointer, val);
	return pointer;
}

inline void init_val(const int dim, int array[], const int val) {
	int i = 0;
	
	for (i = 0; i < dim; i++) {
		array[i] = val;
	}
}

// returns the sum of each entry in the (a+1)-th column of q-matrix (a = 0...(K-1))
inline int sum_q_column(const int a, const int m, const int K, const int q[]) {
	int i = 0, tmp = 0;
	
	for (i=0; i<m; i++) {
		tmp += q[(i*K) + a];
	}
	
	return tmp;
}

// implementing eq. (3.8) of Candelas et.al. (orig. CICY paper)
void split_matrix(FILE * fp, const conf conflist[], const int idx, set<conf> * matrixlist) {
	int i, j, l;
	int subsplit[2*conflist[idx].m];
	int new_m = (conflist[idx].m)+1;
	int new_K = (conflist[idx].K)+1;
	int new_n_r[new_m];
	int new_q[new_m*new_K];
	int new_q_sorted[new_m*new_K];
	int new_eulernumber;

	int detvar_K = (conflist[idx].K)+3;
	int detvar_q[(conflist[idx].m)*detvar_K];
	int detvar_eulernumber = 0;
	vector<jmonomial> mu;

	// initialize new_n_r and new_q
	new_n_r[0] = 1;
	for (i = 0; i < conflist[idx].m; i++) {
		new_n_r[i+1] = conflist[idx].n_r[i];
	}
	new_q[0] = 1;
	new_q[1] = 1;
	for (i = 0; i < (conflist[idx].K-1); i++) {
		new_q[i+2] = 0;
	}

	// walk columns
	for (i = 0; i < conflist[idx].K; i++) {
		if ((i > 0) && (cols_equal(conflist[idx].m, conflist[idx].K, conflist[idx].q, i-1, i) == true)) {
			// don't need to split this column if it is equal to the previous column
			continue;
		}

		// initialize [u,v] part of LHS of eq. (3.8)
		init_val(2*conflist[idx].m, subsplit, 0);
		for (j = 0; j < conflist[idx].m; j++) {
			subsplit[1 + (2*j)] = conflist[idx].q[i + (j*conflist[idx].K)];
		}
		
		// initialize [M] part of LHS of eq. (3.8)
		for (j = 0; j < i; j++) {
			for (l = 0; l < conflist[idx].m; l++) {
				new_q[(2+j) + ((l+1)*new_K)] = conflist[idx].q[j + (l*conflist[idx].K)];
			}
		}
		for (j = i+1; j < conflist[idx].K; j++) {
			for (l = 0; l < conflist[idx].m; l++) {
				new_q[(1+j) + ((l+1)*new_K)] = conflist[idx].q[j + (l*conflist[idx].K)];
			}
		}
		
		do {
			if (is_triv_subsplit(conflist[idx].m, subsplit)) {
				// ^^ check that each column in subsplit adds up to > 0
				continue;
			}
			for (j = 0; j < conflist[idx].m; j++) {
				new_q[new_K*(j+1)] = subsplit[2*j];
				new_q[(new_K*(j+1))+1] = subsplit[(2*j)+1];
			}
			//debug_output(new_m, new_K, new_n_r, new_q);
			//debug_output(conflist[idx].m, 2, conflist[idx].n_r, subsplit); fflush(stdout);
			mymemcpy(new_q_sorted, new_q, new_m*new_K);
			sort_q(new_m, new_K, new_q_sorted);

			// initialize [u,u,v,v] part of determinantal variety
			for (j = 0; j < conflist[idx].m; j++) {
				detvar_q[detvar_K*j] = subsplit[2*j];
				detvar_q[(detvar_K*j)+1] = subsplit[2*j];
				detvar_q[(detvar_K*j)+2] = subsplit[(2*j)+1];
				detvar_q[(detvar_K*j)+3] = subsplit[(2*j)+1];
			}
			// initialize [M] part of determinantal variety
			for (j = 0; j < i; j++) {
				for (l = 0; l < conflist[idx].m; l++) {
					detvar_q[(4+j) + (l*detvar_K)] = conflist[idx].q[j + (l*conflist[idx].K)];
				}
			}
			for (j = i+1; j < conflist[idx].K; j++) {
				for (l = 0; l < conflist[idx].m; l++) {
					detvar_q[(3+j) + (l*detvar_K)] = conflist[idx].q[j + (l*conflist[idx].K)];
				}
			}
			clear_mu(&mu);
			calc_mu(conflist[idx].m, detvar_K, conflist[idx].n_r, detvar_q, &mu);

			detvar_eulernumber = compute_detvar_eulernumber_fast(n-3, conflist[idx].m, detvar_K, conflist[idx].n_r, detvar_q, &mu);
			if (n == 4) {
				//if (detvar_eulernumber != 0) {
				//	// ^^ "detvar_eulernumber != 0" means, we have an eulernumber-changing effective splitting
				//	continue;
				//}
				
				// criterion for ineffective vs. eulernumber-preserving effective splitting:
				//   effective if "detvar_eulernumber == 0" and detvar consists of a configuration of tori
				// ineffective if "detvar_eulernumber == 0" and detvar is an empty set
				
				// debug (NEW 28/02/12)
				//if (is_detvar_emptyset(conflist[idx].m, detvar_K, conflist[idx].n_r, detvar_q, &mu) != is_detvar_emptyset_old(conflist[idx].m, detvar_K, conflist[idx].n_r, detvar_q)) {
				//	debug_output(conflist[idx].m, detvar_K, conflist[idx].n_r, detvar_q);
				//	print_mu(conflist[idx].m, &mu);
				//	printf("chi = %i\n", detvar_eulernumber);
				//}
				if ( (detvar_eulernumber == 0) && (is_detvar_emptyset(conflist[idx].m, detvar_K, conflist[idx].n_r, detvar_q, &mu) == true) ) {
					continue;
				}
				// is_detvar_emptyset == false  <==>	effective splitting
				// is_detvar_emptyset == true  <==> 	ineffective splitting
				//printf("%i\n", is_detvar_emptyset(conflist[idx].m, detvar_K, conflist[idx].n_r, detvar_q));
				//debug_output(conflist[idx].m, detvar_K, conflist[idx].n_r, detvar_q); fflush(stdout);
				
				new_eulernumber = conflist[idx].eulernumber + (3*detvar_eulernumber);
			}
			else if (n == 3) {
				if (detvar_eulernumber == 0) {
					// ^^ if the euler numbers between the old and the new (split) matrix agree, we have an ineffective splitting
					// and the manifolds are equal. Hence we want to exclude the new (split) manifold from the list in this case (to avoid redundancy)
					continue;
				}
				new_eulernumber = conflist[idx].eulernumber + (2*detvar_eulernumber);
			}
			else {
				printf("for n != 3,4 need euler.cpp & mathematica calls.\n");
				return;
			}

			if (is_already_there(matrixlist, new_m, new_K, new_n_r, new_q_sorted) != matrixlist->end()) {
				continue;
			}

			add_conf(matrixlist, new_m, new_K, new_n_r, new_q_sorted, new_eulernumber);
			write_output1(fp, ++new_counter, new_m, new_K, new_n_r, new_q_sorted, new_eulernumber);
			fflush(fp);	// debug  <-- a bit slower (~+5%) if flushed here
		} while(increase_subsplit(conflist[idx].m, subsplit));
	}

	clear_mu(&mu);
}

// returns TRUE if col1 == col2 (treated as columns of q-matrix)
inline int cols_equal(const int m, const int K, const int q[], const int col1, const int col2) {
	int i;

	for (i = 0; i < m; i++) {
		if (q[col1 + (i*K)] != q[col2 + (i*K)]) {
			return false;
		}
	}
	
	return true;
}

// returns false if subsplit-matrix cannot be increased further
int increase_subsplit(const int m, register int q[]) {
	int i = 0, j; register int row;

	for (i=0; i < m; i++) {
		row = i*2;
		if (q[row+1] == 0) {
			// can't increase this row further
			continue;	// "break;" doesn't work!!
		}
		
		// no changes to current q-matrix up to here
		q[row]++;
		q[row+1]--;
		if (i > 0) {
			// if i > 0, then all previous rows must be at their "max"-point, i.e. (x,0), 
			// so to go back to the start point we just need to swap "x" and "0"
			for (j = 0; j < i; j++) {
				q[(j*2)+1] = q[j*2];
				q[j*2] = 0;
			}
		}
		return true;
	}

	return false;
}

// returns TRUE if n_r1[] == n_r2[] (as arrays)
inline int is_equal_ambient_space(const int m, const int n_r1[], const int n_r2[]) {
	int i;

	for (i = 0; i < m; i++) {
		if (n_r1[i] != n_r2[i]) {
			return false;
		}
	}
	
	return true;
}

// returns TRUE if one of the two columns in adds up to zero
inline int is_triv_subsplit(const int m, const int subsplit[]) {
	if ((sum_q_column(0, m, 2, subsplit) == 0) || (sum_q_column(1, m, 2, subsplit) == 0)) {
		return true;
	}
	else {
		return false;
	}
}